# coding:utf-8

"""
@source https://git.code.oa.com/qci/PyTgit
"""
from __future__ import absolute_import
import logging
from urllib.parse import quote
import os.path
import requests
import json
from requests.exceptions import HTTPError

logger = logging.getLogger("tgit")
proxies = {
    "http": "http://dev-proxy.oa.com:8080",
    "https": "http://dev-proxy.oa.com:8080",
}


class TgitClient(object):

    url = "https://git.code.tencent.com/api"

    version = 3

    def __init__(self, auth, **options):
        self.auth = auth
        self.opener = None
        self.options = options

    def build_opener(self):
        if self.opener is None:
            self.opener = requests.Session()
            self.opener.auth = self.auth
        return self.opener

    def to_url(self, _path):
        """
        生成 url
        """
        return "%s/v%s%s" % (self.url, self.version, _path)

    def request(self, method, _path, headers=None, **data):
        """
        发起请求
        """
        url = self.to_url(_path)
        # request_kwargs = dict()
        request_kwargs = {"proxies": proxies}
        if data:
            keys = list(data.keys())
            for key in keys:

                # 参数包含 python 关键字, 添加了一个 '_' 处理, 这里要识别到并移除下划线
                if key.startswith("_"):
                    val = data.pop(key)
                    key = key[1:]
                    request_kwargs[key] = val
                    # set new key
                    # data[key] = val

                placeholder = ":%s" % key
                if placeholder in url:
                    url = url.replace(placeholder, quote(str(data[key]), safe=""))
                    data.pop(key)

        method = method.lower()
        try:
            if not data:
                data = None
            options = {
                "params": data if method in ["get", "delete", "options", "head"] else None,
                "data": data if method in ["post", "put", "patch"] else None,
            }
            options.update(request_kwargs)

            # issue bug
            if options.get("data"):
                options["data"] = json.dumps(options["data"])
            options.update(self.options)
            request = getattr(self.build_opener(), method)
            response = request(url, headers=headers, **options)
            # print(response.text)
            response.raise_for_status()
            if response.status_code == 204:  # no content
                return None
            elif method == "delete":  # delete not read content
                return None
            return response.json()
        except HTTPError as e:
            logger.error("call %s fail" % url)
            raise

    def get(self, _path, **params):
        """
        使用 GET 请求数据
        """
        return self.request("GET", _path, **params)

    def post(self, _path, **data):
        """
        使用 POST 请求数据
        """
        return self.request("POST", _path, **data)

    def put(self, _path, **data):
        """
        使用 PUT 请求数据
        """
        return self.request("PUT", _path, **data)

    def delete(self, _path, **data):
        """
        使用 DELETE 请求数据
        """
        return self.request("DELETE", _path, **data)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/groups.md
    ##
    ##################################################################################

    def create_groups(self, **kwargs):
        """

        新建项目组

        :param name: (string) 项目组的名字
        :param _path: (string) 项目组的路径
        :param description: (string) 关于这个项目组的描述
        """
        return self.post("/groups", **kwargs)

    def update_groups(self, **kwargs):
        """

        编辑项目组

        :param id: (integer) 用户的项目组 ID 或者路径
        :param name: (string) 项目组的名字
        :param description: (string) 关于这个项目组的描述
        """
        return self.put("/groups", **kwargs)

    def delete_groups(self, id, **kwargs):
        """

        删除项目组

        :param id: (integer) id = 项目组唯一标识或路径
        """
        return self.delete("/groups/:id", id=id, **kwargs)

    def list_groups(self, **kwargs):
        """

        获取项目组列表

        :param page: (integer（可选）) 分页 (默认值:1)
        :param per_page: (integer（可选）) 默认页面大小 (默认值 20，最大值： 100)
        """
        return self.get("/groups", **kwargs)

    def search_groups(self, search, **kwargs):
        """

        搜索项目组

        :param page: (integer（可选）) 分页 (默认值:1)
        :param per_page: (integer（可选）) 默认页面大小 (默认值 20，最大值： 100)
        """
        return self.get("/groups", search=search, **kwargs)

    def list_groups_members(self, id, **kwargs):
        """

        获取项目组成员列表

        :param id: (integer) id = 项目组唯一标识或路径
        :param page: (integer（可选）) 分页 (默认:1)
        :param per_page: (integer（可选）) 默认页面大小 (默认 20，最大： 100)
        """
        return self.get("/groups/:id/members", id=id, **kwargs)

    def create_groups_members(self, id, **kwargs):
        """

        增加项目组成员

        :param id: (integer) id = 项目组唯一标识或路径
        :param user_id: (integer) 添加用户的 ID
        :param access_level: (integer) 项目访问级别
        """
        return self.post("/groups/:id/members", id=id, **kwargs)

    def update_groups_members(self, id, user_id, **kwargs):
        """

        修改项目组成员

        :param id: (integer) id = 项目组唯一标识或路径
        :param user_id: (integer) 项目组成员的 ID
        :param access_level: (integer) 项目访问级别
        """
        return self.put("/groups/:id/members/:user_id", id=id, user_id=user_id, **kwargs)

    def delete_groups_members(self, id, user_id, **kwargs):
        """

        移除一个项目组成员

        :param id: (integer) id = 项目组唯一标识或路径
        :param user_id: (integer) 项目组成员的 ID
        """
        return self.delete("/groups/:id/members/:user_id", id=id, user_id=user_id, **kwargs)

    def get_groups(self, id, **kwargs):
        """

        获取项目组的详细信息以及项目组下所有项目

        :param id: (integer) id = 项目组唯一标识或路径
        """
        return self.get("/groups/:id", id=id, **kwargs)

    def create_groups_projects(self, id, project_id, **kwargs):
        """

        把某个项目转让到指定项目组

        :param id: (integer) id = 项目组唯一标识或路径
        :param project_id: (integer) 项目的 ID
        """
        return self.post("/groups/:id/projects/:project_id", id=id, project_id=project_id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/namespaces.md
    ##
    ##################################################################################

    def list_namespaces(self, **kwargs):
        """

        获取命名空间列表

        :param page: (integer （可选）) 分页 (默认值:1)
        :param per_page: (integer （可选）) 默认页面大小 (默认值 20，最大值： 100)
        """
        return self.get("/namespaces", **kwargs)

    def search_namespaces(self, search, **kwargs):
        """

        搜索命名空间

        """
        return self.get("/namespaces", search=search, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/users.md
    ##
    ##################################################################################

    def list_users(self, **kwargs):
        """

        作为管理者

        """
        return self.get("/users", **kwargs)

    def list_users(self, **kwargs):
        """

        作为普通用户

        """
        return self.get("/users", **kwargs)

    def get_user_watched(self, **kwargs):
        """

        获用户关注项目列表

        """
        return self.get("/user/watched", **kwargs)

    def create_users(self, **kwargs):
        """

        新增用户

        :param email: (string) 邮箱地址
        :param password: (string（可选）) 密码
        :param username: (string) 用户
        :param extern_uid: (integer) 外部系统唯一标识
        :param provider: (string) 外部系统认证方式
        :param name: (string (可选)) 用户名称
        :param skype: (string (可选)) Skype ID
        :param linkedin: (string (可选)) LinkedIn
        :param twitter: (string (可选)) Twitter account
        :param qq: (string (可选)) QQ
        :param wechat: (string (可选)) 微信号
        :param website_url: (string (可选)) Website URL
        :param projects_limit: (integer (可选)) 可以创建的项目数
        :param bio: (string (可选)) User's biography
        :param admin: (boolean (可选)) User is admin - true or false (default)
        :param can_create_group: (boolean (可选)) User can create groups - true or false
        :param confirm: (boolean (可选)) Require confirmation - true (default) or false
        """
        return self.post("/users", **kwargs)

    def update_users(self, id, **kwargs):
        """

        修改用户

        :param email: (string) 邮箱
        :param password: (string) 密码
        :param username: (string) 用户名称
        :param name: (string) 用户
        :param skype: (string (可选)) Skype ID
        :param linkedin: (string (可选)) LinkedIn
        :param twitter: (string (可选)) Twitter account
        :param qq: (string (可选)) QQ
        :param wechat: (string (可选)) 微信号
        :param website_url: (string (可选)) Website URL
        :param projects_limit: (integer (可选)) Number of projects user can create
        :param extern_uid: (integer (可选)) External UID
        :param provider: (string (可选)) External provider name
        :param bio: (string (可选)) User's biography
        :param admin: (boolean (可选)) User is admin - true or false (default)
        :param can_create_group: (boolean (可选)) User can create groups - true or false
        """
        return self.put("/users/:id", id=id, **kwargs)

    def delete_users(self, id, **kwargs):
        """

        删除用户

        :param id: (integer) id=用户唯一标识或用户名称
        """
        return self.delete("/users/:id", id=id, **kwargs)

    def get_users(self, id, **kwargs):
        """

        获取某个用户的账号信息

        :param id: (integer 或 string) id=用户唯一标识或用户名称
        """
        return self.get("/users/:id", id=id, **kwargs)

    def get_user(self, **kwargs):
        """

        当前认证用户

        """
        return self.get("/user", **kwargs)

    def create_user_keys(self, **kwargs):
        """

        给当前用户创建一个SSH key

        :param title: (string) SSH Key 的标题
        :param key: (string) 新的 SSH key
        """
        return self.post("/user/keys", **kwargs)

    def create_users_keys(self, id, **kwargs):
        """

        给指定用户创建一个 SSH key

        :param id: (integer) id=用户唯一标识或用户名称
        :param title: (string) SSH Key 的标题
        :param key: (string) 新的 SSH key
        """
        return self.post("/users/:id/keys", id=id, **kwargs)

    def list_user_keys(self, **kwargs):
        """

        获取当前用户的SSH key

        """
        return self.get("/user/keys", **kwargs)

    def get_user_keys(self, id, **kwargs):
        """

        获取某个指定的 SSH key

        :param id: (integer) SSH key 的 id
        """
        return self.get("/user/keys/:id", id=id, **kwargs)

    def list_users_keys(self, uid, **kwargs):
        """

        获取某个用户的SSH key

        :param uid: (integer) uid=用户唯一标识
        """
        return self.get("/users/:uid/keys", uid=uid, **kwargs)

    def delete_user_keys(self, id, **kwargs):
        """

        删除当前用户的 SSH key

        :param id: (integer) SSH key 的 ID
        """
        return self.delete("/user/keys/:id", id=id, **kwargs)

    def delete_users_keys(self, uid, id, **kwargs):
        """

        删除指定用户的 SSH key

        :param uid: (integer) uid=用户唯一标识
        :param id: (integer) SSH key 的 ID
        """
        return self.delete("/users/:uid/keys/:id", uid=uid, id=id, **kwargs)

    def create_user_emails(self, **kwargs):
        """

        添加邮箱

        :param email: (string) 邮箱地址
        """
        return self.post("/user/emails", **kwargs)

    def create_users_emails(self, id, **kwargs):
        """

        给指定用户添加邮箱

        :param id: (integer) id=用户唯一标识或用户名称
        :param email: (string) 邮箱地址
        """
        return self.post("/users/:id/emails", id=id, **kwargs)

    def get_user_email(self, **kwargs):
        """

        通过邮箱获取用户信息

        :param email: (string) 邮箱地址
        """
        return self.get("/user/email", **kwargs)

    def list_user_emails(self, **kwargs):
        """

        获取用户邮箱列表

        """
        return self.get("/user/emails", **kwargs)

    def list_users_emails(self, uid, **kwargs):
        """

        获取指定用户邮箱列表

        :param uid: (integer) uid=用户唯一标识
        """
        return self.get("/users/:uid/emails", uid=uid, **kwargs)

    def get_user_emails(self, id, **kwargs):
        """

        获取邮箱信息

        :param id: (integer) 邮箱的 ID
        """
        return self.get("/user/emails/:id", id=id, **kwargs)

    def delete_user_emails(self, id, **kwargs):
        """

        删除当前用户的邮箱

        :param id: (integer) 邮箱的 ID
        """
        return self.delete("/user/emails/:id", id=id, **kwargs)

    def delete_users_emails(self, uid, id, **kwargs):
        """

        删除指定用户的邮箱

        :param uid: (integer) id=用户唯一标识或用户名称
        :param id: (integer) 邮箱的 ID
        """
        return self.delete("/users/:uid/emails/:id", uid=uid, id=id, **kwargs)

    def update_users_block(self, uid, **kwargs):
        """

        锁住用户

        :param uid: (integer) uid=用户唯一标识
        """
        return self.put("/users/:uid/block", uid=uid, **kwargs)

    def update_users_unblock(self, uid, **kwargs):
        """

        取消锁住用户

        :param uid: (integer) uid=用户唯一标识
        """
        return self.put("/users/:uid/unblock", uid=uid, **kwargs)

    def create_users_identity(self, uid, **kwargs):
        """

        创建认证方式

        :param uid: (integer) uid=用户唯一标识
        :param extern_uid: (integer) 外部系统唯一标识
        :param provider: (string) 外部系统认证方式
        """
        return self.post("/users/:uid/identity", uid=uid, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/issues.md
    ##
    ##################################################################################

    def create_projects_issues(self, id, **kwargs):
        """

        新建缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param title: (string) 缺陷标题
        :param description: (string（可选）) 缺陷描述
        :param assignee_ids: (string（可选）) 处理人唯一标识，允许多个，以,分隔，最多：10
        :param milestone_id: (integer（可选）) 里程碑 唯一标识
        :param labels: (string （可选）) 缺陷标签，允许多个，以,分隔，最多：10
        """
        # data['private_token']='NY2VjIH7EcxmuXW0elkX'
        # result = requests.session().post("https://git.code.tencent.com/api/v3/projects/116331/issues",data,proxies=proxies)
        # return result
        return self.post("/projects/:id/issues", id=id, **kwargs)

    def update_projects_issues(self, id, issue_id, **kwargs):
        """

        修改缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷唯一标识
        :param title: (string（可选）) 缺陷标题
        :param description: (string（可选）) 缺陷描述
        :param assignee_ids: (string（可选）) 处理人唯一标识，允许多个，以,分隔，最大值：10
        :param milestone_id: (integer（可选）) 里程碑的唯一标识
        :param labels: (string （可选）) 缺陷标签，允许多个，以,分隔，最大值：10
        :param state_event: (string （可选）) 缺陷事件，可选reopen、close
        """
        return self.put("/projects/:id/issues/:issue_id", id=id, issue_id=issue_id, **kwargs)

    def list_issues(self, **kwargs):
        """

        用户创建缺陷列表

        :param state: (string （可选）) 缺陷状态，可选opened, closed
        :param labels: (string （可选）) 标签，允许多个，以,分隔，最大值：10
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        :param order_by: (string （可选）) 排序字段，允许按 created_at,updated_at排序（默认created_at）
        :param sort: (string （可选）) 排序方式，允许 asc,desc（默认desc）
        """
        return self.get("/issues", **kwargs)

    def list_projects_issues(self, id, **kwargs):
        """

        项目缺陷列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param iid: (integer （可选）) 缺陷的项目编号
        :param state: (string （可选）) 缺陷状态，可选opened, closed
        :param labels: (string （可选）) 标签，允许多个，以,分隔，最多：10
        :param milestone: (string （可选）) 里程碑标题
        :param order_by: (string （可选）) 排序字段，允许按created_at,updated_at排序（默认created_at）
        :param sort: (string （可选）) 排序方式，允许asc,desc（默认desc）
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/:id/issues", id=id, **kwargs)

    def get_projects_issues(self, id, **kwargs):
        """

        查看指定缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷唯一标识
        """
        return self.get("/projects/:id/issues", id=id, **kwargs)

    def get_projects_issue(self, id, issue_id, **kwargs):
        """

        查看指定缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷唯一标识
        """
        return self.get("/projects/:id/issues/:issue_id", id=id, issue_id=issue_id, **kwargs)

    def get_projects_issues_subscribe(self, id, issue_id, **kwargs):
        """

        判断是否订阅给定项目的某个缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷 唯一标识
        """
        return self.get(
            "/projects/:id/issues/:issue_id/subscribe", id=id, issue_id=issue_id, **kwargs
        )

    def update_projects_issues_subscribe(self, id, issue_id, **kwargs):
        """

        订阅给定项目的某个缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷唯一标识
        """
        return self.put(
            "/projects/:id/issues/:issue_id/subscribe", id=id, issue_id=issue_id, **kwargs
        )

    def update_projects_issues_unsubscribe(self, id, issue_id, **kwargs):
        """

        取消订阅给定项目的某个缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷唯一标识
        """
        return self.put(
            "/projects/:id/issues/:issue_id/unsubscribe", id=id, issue_id=issue_id, **kwargs
        )

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/labels.md
    ##
    ##################################################################################

    def create_projects_labels(self, id, **kwargs):
        """

        新增标签

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param name: (string) 标签名
        :param color: (string) 标签颜色，举例：#428bca
        """
        return self.post("/projects/:id/labels", id=id, **kwargs)

    def update_projects_labels(self, id, **kwargs):
        """

        修改标签

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param name: (string) 旧标签名
        :param new_name: (string) 新标签名
        :param color: (string) 标签颜色，举例：#428bca
        """
        return self.put("/projects/:id/labels", id=id, **kwargs)

    def list_projects_labels(self, id, **kwargs):
        """

        获取标签列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param order_by: (string（可选）) 排序字段，允许按 name,created_at排序（默认name）
        :param sort: (string（可选）) 排序方式，允许 asc,desc（默认asc）
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/:id/labels", id=id, **kwargs)

    def delete_projects_labels(self, id, **kwargs):
        """

        删除标签

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param name: (string) 标签名
        """
        return self.delete("/projects/:id/labels", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/milestones.md
    ##
    ##################################################################################

    def create_projects_milestones(self, id, **kwargs):
        """

        新增里程碑

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param title: (string) 里程碑的标题
        :param description: (string（可选）) 里程碑的描述
        :param due_date: (datetime（可选）) 到期时间
        """
        return self.post("/projects/:id/milestones", id=id, **kwargs)

    def update_projects_milestones(self, id, milestone_id, **kwargs):
        """

        编辑里程碑

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param milestone_id: (integer) 里程碑的唯一标识
        :param title: (string) 里程碑的标题
        :param description: (string（可选）) 里程碑的描述
        :param due_date: (datetime（可选）) 到期时间
        :param state_event: (string （可选）) 里程碑的事件，可选active、close
        """
        return self.put(
            "/projects/:id/milestones/:milestone_id", id=id, milestone_id=milestone_id, **kwargs
        )

    def get_projects_milestone(self, id, milestone_id, **kwargs):
        """

        返回指定里程碑

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param milestone_id: (integer) 里程碑的唯一标识
        """
        return self.get(
            "/projects/:id/milestones/:milestone_id", id=id, milestone_id=milestone_id, **kwargs
        )

    def get_projects_milestones(self, id, **kwargs):
        """

        返回指定里程碑

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param milestone_id: (integer) 里程碑的唯一标识
        """
        return self.get("/projects/:id/milestones", id=id, **kwargs)

    def delete_projects_milestones(self, id, milestone_id, **kwargs):
        """

        删除里程碑

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param milestone_id: (integer) 里程碑的唯一标识
        """
        return self.delete(
            "/projects/:id/milestones/:milestone_id", id=id, milestone_id=milestone_id, **kwargs
        )

    def list_projects_milestones_issues(self, id, milestone_id, **kwargs):
        """

        返回某个里程碑下的所有缺陷

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param milestone_id: (integer) 里程碑的唯一标识
        """
        return self.get(
            "/projects/:id/milestones/:milestone_id/issues",
            id=id,
            milestone_id=milestone_id,
            **kwargs
        )

    def list_projects_milestones(self, id, **kwargs):
        """

        返回里程碑列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param iid: (integer （可选）) iid
        :param state: (string （可选）) 里程碑的状态，可选active, closed
        :param order_by: (string （可选）) 排序字段， 允许按id,due_date,created_at,updated_at排序（默认created_at）
        :param sort: (string （可选）) 排序方式， 允许 asc,desc（默认desc）
        :param page: (integer（可选）) 分页（default：1）
        :param per_page: (integer（可选）) 默认页面大小（default：20，max：100）
        """
        return self.get("/projects/:id/milestones", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/watchers.md
    ##
    ##################################################################################

    def list_projects_watchers(self, id, **kwargs):
        """

        关注者列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/:id/watchers", id=id, **kwargs)

    def get_projects_watch(self, id, **kwargs):
        """

        是否关注给定项目

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.get("/projects/:id/watch", id=id, **kwargs)

    def update_projects_watch(self, id, **kwargs):
        """

        关注项目

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param mute: (boolean) 是否静音，mute = true将不会收到除参与和订阅以外通知，默认mute = false
        """
        return self.put("/projects/:id/watch", id=id, **kwargs)

    def delete_projects_watch(self, id, **kwargs):
        """

        取消关注项目

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.delete("/projects/:id/watch", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/repositorys.md
    ##
    ##################################################################################

    def get_projects_repository_archive(self, id, **kwargs):
        """

        下载存档的项目版本库

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string（可选）) commit hash值、分支名或 tag
        """
        return self.get("/projects/:id/repository/archive", id=id, **kwargs)

    def list_projects_repository_contributors(self, id, **kwargs):
        """

        获取贡献者列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.get("/projects/:id/repository/contributors", id=id, **kwargs)

    def get_projects_repository_blobs(self, id, sha, **kwargs):
        """

        获取原始文件内容

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        :param file_path: (string) 文件路径 (文件名）
        """
        return self.get("/projects/:id/repository/blobs/:sha", id=id, sha=sha, **kwargs)

    def get_projects_repository_raw_blobs(self, id, sha, **kwargs):
        """

        获取 blob 原始内容

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        """
        return self.get("/projects/:id/repository/raw_blobs/:sha", id=id, sha=sha, **kwargs)

    def get_projects_repository_compare(self, id, _from, to, straight, **kwargs):
        """

        获取差异内容

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param from: (string) 提交的 hash、分支 或 tag
        :param to: (string) 提交的 hash、分支 或 tag
        :param straight: (boolean（可选）) true：两个点比较差异，false：三个点比较差异。默认是 false
        """
        return self.get(
            "/projects/:id/repository/compare",
            id=id,
            _from=_from,
            to=to,
            straight=straight,
            **kwargs
        )

    def get_projects_repository_compare_commits(self, id, **kwargs):
        """

        获取差异commits
        http://x.code.oa.com/Code/Git/blob/master/help/api/repositorys.md#%E4%B8%8B%E8%BD%BDcompare%E5%B7%AE%E5%BC%82%E6%96%87%E4%BB%B6%E9%9B%86
        """
        return self.get("/projects/:id/repository/compare/commits", id=id, **kwargs)

    def get_projects_repository_tree(self, id, **kwargs):
        """

        获取版本库文件和目录列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param ref_name: (string（可选）) commit hash值、分支 或 tag，默认：默认分支
        :param _path: (string（可选）) 文件路径
        """
        return self.get("/projects/:id/repository/tree", id=id, **kwargs)

    def list_projects_repository_files(self, id, **kwargs):
        """

        获取单个文件内容和信息

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param ref: (string) ref可以是SHA, branch name, tag name
        :param file___path: (string) 文件路径（文件名）
        """
        return self.get("/projects/:id/repository/files", id=id, **kwargs)

    def create_projects_repository_files(self, id, **kwargs):
        """

        新增文件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param file___path: (string) 文件路径(文件名)
        :param branch_name: (string) 分支 名
        :param encoding: (string（可选）) 内容编码，可选：text、base64，默认：text
        :param content: (string) 内容
        :param commit_message: (string) 描述信息
        """
        return self.post("/projects/:id/repository/files", id=id, **kwargs)

    def upload_images(self, id, image_path):
        """

        上传附件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param file_path: (string) 文件路径(文件名)
        """
        return self.post(
            "/projects/:id/uploads",
            id=id,
            _files={"file": ("file", open(image_path, "rb"), "image/png")},
        )

    def upload_file(self, id, path, content_type):
        """

        上传附件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param file_path: (string) 文件路径(文件名)
        """
        return self.post(
            "/projects/:id/uploads",
            id=id,
            _files={"file": (os.path.basename(path), open(path, "rb"), content_type)},
        )

    def delete_projects_repository_files(self, id, **kwargs):
        """

        删除文件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param file___path: (string) 文件路径（文件名）
        :param branch_name: (string) 分支名
        :param commit_message: (string) 描述
        """
        return self.delete("/projects/:id/repository/files", id=id, **kwargs)

    def update_projects_repository_files(self, id, **kwargs):
        """

        编辑文件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param file___path: (string) 文件路径（文件名）
        :param branch_name: (string) 分支名
        :param encoding: (string（可选）) 内容编码，可选：text、base64，默认：text
        :param content: (string) 内容
        :param commit_message: (string) 注释
        """
        return self.put("/projects/:id/repository/files", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/branches.md
    ##
    ##################################################################################

    def create_projects_repository_branches(self, id, **kwargs):
        """

        创建项目分支

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch_name: (string) 分支名
        :param ref: (string) ref可以是已存在的SHA, branch name, tag name
        """
        return self.post("/projects/:id/repository/branches", id=id, **kwargs)

    def delete_projects_repository_branches(self, id, branch, **kwargs):
        """

        删除分支

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch: (string) 分支 名
        """
        return self.delete(
            "/projects/:id/repository/branches/:branch", id=id, branch=branch, **kwargs
        )

    def list_projects_repository_branches(self, id, **kwargs):
        """

        分支列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认：20，最大：100）
        """
        return self.get("/projects/:id/repository/branches", id=id, **kwargs)

    def get_projects_repository_branches(self, id, branch, **kwargs):
        """

        获取分支详情

        :param id: (integer 或string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch: (string) 分支名
        """
        return self.get("/projects/:id/repository/branches/:branch", id=id, branch=branch, **kwargs)

    def update_projects_repository_branches_protect(self, id, branch, **kwargs):
        """

        将分支变为保护分支

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch: (string) 分支 名
        :param developers_can_push: (boolean（可选）) developer角色是否能够 推送
        :param developers_can_merge: (boolean（可选）) developer角色是否能够 合并
        :param push_access_level: (integer（可选）) 推送权限级别，有效值：0，30，40。0 ：Owner + Only Branch Members(Except Reporters);30：Above Developer + Branch Members(Except Reporters);40：Above Master + Branch Members(Except Reporters)
        :param merge_access_level: (integer（可选）) 合并权限级别，有效值：0，30，40。0 ：Owner + Only Branch Members(Except Reporters);30：Above Developer + Branch Members(Except Reporters);40：Above Master + Branch Members(Except Reporters)
        """
        return self.put(
            "/projects/:id/repository/branches/:branch/protect", id=id, branch=branch, **kwargs
        )

    def update_projects_repository_branches_unprotect(self, id, branch, **kwargs):
        """

        取消保护分支

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch: (string) 分支名
        """
        return self.put(
            "/projects/:id/repository/branches/:branch/unprotect", id=id, branch=branch, **kwargs
        )

    def create_projects_branches_protected_members(self, id, branch_name, **kwargs):
        """

        为保护分支添加成员

        :param id: (integer) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param protected_branch_name: (integer) 保护分支的 分支名
        :param user_id: (integer) 用户的 ID
        :param access_level: (integer) 要给用户设置的访问级别
        """
        return self.post(
            "/projects/:id/branches/protected/:branch_name/members",
            id=id,
            branch_name=branch_name,
            **kwargs
        )

    def list_projects_branches_protected_members(self, id, protected_branch_name, **kwargs):
        """

        获取保护分支的成员

        :param id: (integer) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param protected_branch_name: (integer) 保护分支的 分支名
        """
        return self.get(
            "/projects/:id/branches/protected/:protected_branch_name/members",
            id=id,
            protected_branch_name=protected_branch_name,
            **kwargs
        )

    def update_projects_branches_protected_members(
        self, id, protected_branch_name, user_id, **kwargs
    ):
        """

        编辑保护分支的成员

        :param id: (integer) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param protected_branch_name: (integer) 保护分支的 分支名
        :param user_id: (integer) 用户的 ID
        :param access_level: (integer) 要给用户设置的访问级别
        """
        return self.put(
            "/projects/:id/branches/protected/:protected_branch_name/members/:user_id",
            id=id,
            protected_branch_name=protected_branch_name,
            user_id=user_id,
            **kwargs
        )

    def delete_projects_branches_protected_members(
        self, id, protected_branch_name, user_id, **kwargs
    ):
        """

        删除保护分支的指定用户

        :param id: (integer) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param protected_branch_name: (integer) 保护分支的 分支名
        :param user_id: (integer) 用户的 ID
        """
        return self.delete(
            "/projects/:id/branches/protected/:protected_branch_name/members/:user_id",
            id=id,
            protected_branch_name=protected_branch_name,
            user_id=user_id,
            **kwargs
        )

    def get_projects_tloc_branch_lifecycle(self, id, **kwargs):
        """

        查询分支生命周期

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        """
        return self.get("/projects/:id/tloc/branch/lifecycle", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/releases.md
    ##
    ##################################################################################

    def list_projects_releases(self, id, **kwargs):
        """

        获取releases列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param labels: (string （可选）) 发布标签， 允许多个，以,分隔，最多：10
        :param author_id: (integer（可选）) 创建人id
        :param type: (string （可选）) 发布类型，目前只有 release
        :param page: (integer（可选）) 分页（default：1）
        :param per_page: (integer（可选）) 默认页面大小（default：20，max：100）
        """
        return self.get("/projects/:id/releases", id=id, **kwargs)

    def get_projects_releases(self, id, tag_name, **kwargs):
        """

        获取某个指定的release

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (String) release的tag字段
        """
        return self.get("/projects/:id/releases/:tag_name", id=id, tag_name=tag_name, **kwargs)

    def create_projects_releases(self, id, **kwargs):
        """

        新增一个release

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag: (string) tag 名
        :param start_point: (string) 从 commit hash、存在的 branch 或 tag 创建 releases
        :param tag_message: (string（可选）) 对tag的描述、注释
        :param title: (string（可选）) 对releases的标题，如果为空则取tag名
        :param type: (string（可选）) 发布类型， 默认release (type为release的发布将会出现在版本发布列表中，如果不需要则传空字符串" ")
        :param description: (string（可选）) 对release的描述
        :param labels: (string（可选）) 发布标签， 允许多个，以,分隔，最多：10
        """
        return self.post("/projects/:id/releases", id=id, **kwargs)

    def update_projects_releases(self, id, tag_name, **kwargs):
        """

        更新一个release

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (string) tag名
        :param type: (string（可选）) 发布类型， 默认release (type为release的发布将会出现在版本发布列表中，如果不需要则传空字符串" ")
        :param description: (string（可选）) 对release的描述
        :param labels: (string（可选）) 发布标签， 允许多个，以,分隔，最多：10
        """
        return self.put("/projects/:id/releases/:tag_name", id=id, tag_name=tag_name, **kwargs)

    def delete_projects_releases(self, id, tag_name, **kwargs):
        """

        删除releases

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (string) tag名
        """
        return self.delete("/projects/:id/releases/:tag_name", id=id, tag_name=tag_name, **kwargs)

    def create_projects_releases_attachments(self, id, tag_name, **kwargs):
        """

        上传附件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (string) tag名
        :param file: (-) 单个附件
        """
        return self.post(
            "/projects/:id/releases/:tag_name/attachments", id=id, tag_name=tag_name, **kwargs
        )

    def get_projects_releases_attachments(self, id, tag_name, attachment_name, **kwargs):
        """

        下载附件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (string) tag名
        :param attachment_name: (string) 附件名
        """
        return self.get(
            "/projects/:id/releases/:tag_name/attachments/:attachment_name",
            id=id,
            tag_name=tag_name,
            attachment_name=attachment_name,
            **kwargs
        )

    def delete_projects_releases_attachments(self, id, tag_name, attachment_name, **kwargs):
        """

        删除附件

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param tag_name: (string) tag名
        :param attachment_name: (string) 附件名
        """
        return self.delete(
            "/projects/:id/releases/:tag_name/attachments/:attachment_name",
            id=id,
            tag_name=tag_name,
            attachment_name=attachment_name,
            **kwargs
        )

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/commits.md
    ##
    ##################################################################################

    def get_projects_repository_commits(self, id, sha, **kwargs):
        """

        取得一个提交

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        """
        return self.get("/projects/:id/repository/commits/:sha", id=id, sha=sha, **kwargs)

    def get_projects_repository_commits_diff(self, id, sha, **kwargs):
        """

        取得提交的差异

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        :param _path: (string（可选）) 文件路径
        :param ignore_white_space: (boolean（可选）) 有差异的内容是否忽略空白符，默认不忽略
        """
        return self.get("/projects/:id/repository/commits/:sha/diff", id=id, sha=sha, **kwargs)

    def create_projects_repository_commits_comments(self, id, sha, **kwargs):
        """

        对一个提交发表评论

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        :param note: (string) 评论内容
        :param _path: (string（可选）) 文件路径
        :param line: (integer（可选）) 行号
        :param line_type: (string（可选）) 变更类型，可选old、new
        """
        return self.post("/projects/:id/repository/commits/:sha/comments", id=id, sha=sha, **kwargs)

    def list_projects_repository_commits_comments(self, id, sha, **kwargs):
        """

        取得一个提交的评论

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        """
        return self.get("/projects/:id/repository/commits/:sha/comments", id=id, sha=sha, **kwargs)

    def list_projects_repository_commits(self, id, **kwargs):
        """

        列出项目版本库所有的提交

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param ref_name: (string（可选）) 版本库分支 或 tag，默认：默认分支
        :param _path: (string（可选）) 文件路径
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/repository/commits", id=id, **kwargs)

    def list_projects_repository_commits_refs(self, id, sha, **kwargs):
        """

        获取某个提交对应的分支和tag

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        :param type: (string（可选）) branch、tag、或all （默认：all）
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/repository/commits/:sha/refs", id=id, sha=sha, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/projects.md
    ##
    ##################################################################################

    def list_projects_members(self, id, **kwargs):
        """

        获取项目成员列表

        :param page: (integer) 页数 (默认值:1)
        :param per_page: (integer) 每页列出成员数 (默认值 20)
        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param query: (string (可选)) 搜索成员的字符串
        """
        return self.get("/projects/:id/members", id=id, **kwargs)

    def create_projects_members(self, id, **kwargs):
        """

        增加项目成员

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param user_id: (integer) 增加的用户的 ID
        :param access_level: (integer) 项目访问级别
        """
        return self.post("/projects/:id/members", id=id, **kwargs)

    def create_projects_members(self, id, **kwargs):
        """ """
        return self.post("/projects/:id/members", id=id, **kwargs)

    def update_projects_members(self, id, user_id, **kwargs):
        """

        修改项目成员

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param user_id: (integer) 用户的 ID
        :param access_level: (integer) 项目访问级别
        """
        return self.put("/projects/:id/members/:user_id", id=id, user_id=user_id, **kwargs)

    def delete_projects_members(self, id, user_id, **kwargs):
        """

        删除项目成员

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param user_id: (integer) 用户的 ID
        """
        return self.delete("/projects/:id/members/:user_id", id=id, user_id=user_id, **kwargs)

    def get_projects_members(self, id, user_id, **kwargs):
        """

        获取项目内的某个指定成员信息

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param user_id: (integer) 成员的 ID
        """
        return self.get("/projects/:id/members/:user_id", id=id, user_id=user_id, **kwargs)

    def create_projects(self, **kwargs):
        """

        创建项目

        :param name: (string) 项目名
        :param _path: (string（可选）) 项目版本库路径，默认：_path = name
        :param fork_enabled: (boolean（可选）) 项目是否可以被fork, 默认:false
        :param namespace_id: (integer （可选）) 项目所属命名空间 ，默认用户的命名空间
        :param description: (string（可选）) 项目描述
        :param visibility_level: (integer（可选）) 项目可视范围，默认visibility_level = 0
        """
        return self.post("/projects", **kwargs)

    def update_projects(self, id, **kwargs):
        """

        编辑项目

        :param name: (string（可选）) 项目名
        :param fork_enabled: (boolean（可选）) 项目是否可以被fork, 默认:false
        :param description: (string（可选）) 项目描述
        :param default_branch: (string（可选）) 项目默认分支
        :param limit_file_size: (float（可选）) 文件大小限制，单位:MB
        :param limit_lfs_file_size: (float（可选）) LFS 文件大小限制，单位:MB
        :param issues_enabled: (boolean（可选）) 缺陷配置
        :param merge_requests_enabled: (boolean（可选）) 合并请求配置
        :param wiki_enabled: (boolean（可选）) 维基配置
        :param review_enabled: (boolean（可选）) 评审配置
        :param tag_name_regex: (string（可选）) 推送或创建 tag 规则
        :param tag_create_push_level: (integer（可选）) 推送或创建 tag 权限
        :param visibility_level: (integer（可选）) 项目可视范围
        """
        return self.put("/projects/:id", id=id, **kwargs)

    def create_projects_user(self, user_id, **kwargs):
        """

        创建属于某个指定用户的项目

        :param user_id: (integer) 项目所属用户
        :param name: (string) 项目名
        :param _path: (string（可选）) 项目版本库路径，默认：_path = name
        :param description: (string（可选）) 项目描述
        :param visibility_level: (integer（可选）) 项目可视范围，默认visibility_level = 0
        """
        return self.post("/projects/user/:user_id", user_id=user_id, **kwargs)

    def list_projects(self, **kwargs):
        """

        获取项目列表

        :param search: (string （可选）) 搜索条件，模糊匹配_path,name
        :param archived: (boolean（可选）) 归档状态，archived = true限制为查询归档项目，默认不区分归档状态
        :param order_by: (string （可选）) 排序字段，允许按 id,name,_path,created_at,updated_at,last_activity_at排序（默认created_at）
        :param sort: (string （可选）) 排序方式，允许asc,desc（默认 desc）
        :param page: (integer（可选）) 页数（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects", **kwargs)

    def get_projects_owned(self, **kwargs):
        """

        获取用户拥有的项目列表

        :param search: (string（可选）) 搜索条件，模糊匹配_path, name
        :param archived: (boolean（可选）) 归档状态，archived = true限制为查询归档项目，默认不区分归档状态
        :param order_by: (string（可选）) 排序字段，允许按 id,name,_path,created_at,updated_at,last_activity_at排序（默认created_at）
        :param sort: (string（可选）) 排序方式，允许 asc or desc（默认 desc）
        :param page: (integer（可选）) 页数（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/owned", **kwargs)

    def get_projects_public(self, **kwargs):
        """

        获取所有公有项目列表

        :param order_by: (string（可选）) 排序字段，允许按 id,name,_path,created_at,updated_at,last_activity_at排序（默认created_at）
        :param sort: (string（可选）) 排序方式，允许 asc or desc（默认 desc）
        :param page: (integer（可选）) 页数（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/public", **kwargs)

    def get_projects_inner_source(self, **kwargs):
        """

        获取内源项目列表

        :param search: (string （可选）) 搜索条件，模糊匹配_path,name
        :param order_by: (string （可选）) 排序字段，允许按 id,name,_path,created_at,updated_at,last_activity_at排序（默认created_at）
        :param sort: (string （可选）) 排序方式，允许asc,desc（默认 desc）
        :param page: (integer（可选）) 页数（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值：20，最大值：100）
        """
        return self.get("/projects/inner_source", **kwargs)

    def delete_projects(self, id, **kwargs):
        """

        删除项目

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.delete("/projects/:id", id=id, **kwargs)

    def get_projects(self, id, **kwargs):
        """

        获取项目详细信息

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.get("/projects/:id", id=id, **kwargs)

    def list_projects_shares(self, id, **kwargs):
        """

        获取项目的share group列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.get("/projects/:id/shares", id=id, **kwargs)

    def create_projects_share(self, id, **kwargs):
        """

        与组共享项目

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param group_id: (integer) 要与之共享的组的id
        :param group_access: (integer) 授予组的权限级别
        """
        return self.post("/projects/:id/share", id=id, **kwargs)

    def delete_projects_share(self, id, group_id, **kwargs):
        """

        删除组中共享项目链接

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param group_id: (integer) 组的id
        """
        return self.delete("/projects/:id/share/:group_id", id=id, group_id=group_id, **kwargs)

    def list_projects(self, **kwargs):
        """

        通过名称搜索项目

        :param search: (string) 搜索条件，模糊匹配_path, name.
        :param with_archived: (boolean（可选）) 归档状态，with_archived = true限制为查询归档项目，默认不区分
        :param with_push: (boolean（可选）) 推送状态，with_push = true限制为查询推送过的项目，默认不区分
        :param abandoned: (boolean（可选）) 活跃状态，abandoned = true限制为查询最近半年更新过的项目，默认全部
        :param visibility_levels: (string（可选）) 项目可视范围，默认visibility_levels = "0, 10, 20"
        :param order_by: (string（可选）) 排序字段，允许按id,name,_path,created_at,updated_at,last_activity_at排序（默认created_at）
        :param sort: (string（可选）) 排序方式，允许asc,desc（默认 desc）
        :param page: (integer（可选）) 页数（默认:1）
        :param per_page: (integer（可选）) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects", **kwargs)

    def list_projects_events(self, id, **kwargs):
        """

        查询项目的事件列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param user_id_or_name: (integer  或 string（可选）) 用户的 ID或用户名
        :param page: (integer) 页数（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/events", id=id, **kwargs)

    def update_projects_star(self, id, **kwargs):
        """

        对指定项目标星

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.put("/projects/:id/star", id=id, **kwargs)

    def delete_projects_star(self, id, **kwargs):
        """

        取消对指定项目标星

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.delete("/projects/:id/star", id=id, **kwargs)

    def get_projects_star(self, id, **kwargs):
        """

        查看对指定项目是否标星

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.get("/projects/:id/star", id=id, **kwargs)

    def list_projects_stars(self, id, **kwargs):
        """

        获取标星项目列表

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param page: (integer（可选）) 页数（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/stars", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/fork.md
    ##
    ##################################################################################

    def create_projects_fork(self, id, **kwargs):
        """

        Fork 项目

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.post("/projects/fork/:id", id=id, **kwargs)

    def create_projects_fork(self, id, forked_from_id, **kwargs):
        """

        建立项目之间的 fork 关系

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param forked_from_id: (integer) 需要 fork 的项目 id
        """
        return self.post(
            "/projects/:id/fork/:forked_from_id", id=id, forked_from_id=forked_from_id, **kwargs
        )

    def delete_projects_fork(self, id, **kwargs):
        """

        删除 Fork 关系

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        """
        return self.delete("/projects/:id/fork", id=id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/commit_check.md
    ##
    ##################################################################################

    def create_projects_commit_statuses(self, id, sha, **kwargs):
        """

        新建检测结果

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值、分支名或 tag
        :param state: (string) 检测状态，可选值：pending, success, error, failure
        :param target_url: (string（可选）) 检测路径，最大字节长度：255
        :param description: (string（可选）) 检测结果描述，最大字节长度：255
        :param context: (string（可选）) 区别于其他检测系统的标签，默认：default
        :param block: (boolean（可选）) 是否锁住提交和合并请求，默认：false
        """
        return self.post("/projects/:id/commit/:sha/statuses", id=id, sha=sha, **kwargs)

    def list_projects_commits_status(self, id, ref, **kwargs):
        """

        通过 Ref 查询特定的组合检测结果

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param ref: (string) ref可以是SHA, branch name, tag name
        :param page: (integer) 分页（默认：1）
        :param per_page: (integer) 默认页面大小（默认： 20，最大： 100）
        """
        return self.get("/projects/:id/commits/:ref/status", id=id, ref=ref, **kwargs)

    def list_projects_commits_statuses(self, id, ref, **kwargs):
        """

        通过 Ref 查询检测结果

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param ref: (string) ref可以是SHA, branch name, tag name
        :param page: (integer) 分页（默认：1）
        :param per_page: (integer) 默认页面大小（默认： 20，最大： 100）
        """
        return self.get("/projects/:id/commits/:ref/statuses", id=id, ref=ref, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/hooks.md
    ##
    ##################################################################################

    def create_projects_hooks(self, id, **kwargs):
        """

        给项目增加回调钩子

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param url: (string) 回调钩子 地址
        :param push_events: (boolean) 有推送事件触发回调钩子，默认 true
        :param issues_events: (boolean) 有缺陷事件触发回调钩子，默认 false
        :param merge_requests_events: (boolean) 有合并请求事件触发回调钩子，默认 false
        :param tag_push_events: (boolean) 有 Tag 推送事件触发回调钩子，默认 false
        :param note_events: (boolean) 有评论事件触发回调钩子，默认 false
        :param review_events: (boolean) 有评审事件触发回调钩子，默认 false
        :param token: (string(可选)) 用以校验收到的负载;此token不会在包含在返回值中
        """
        return self.post("/projects/:id/hooks", id=id, **kwargs)

    def update_projects_hooks(self, id, hook_id, **kwargs):
        """

        编辑项目回调钩子

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param url: (string) 回调钩子 地址
        :param push_events: (boolean（可选）) 有推送事件触发回调钩子
        :param issues_events: (boolean（可选）) 有缺陷事件触发回调钩子
        :param merge_requests_events: (boolean（可选）) 有合并请求事件触发回调钩子
        :param tag_push_events: (boolean（可选）) 有 Tag 推送事件触发回调钩子
        :param note_events: (boolean（可选）) 有评论事件触发回调钩子
        :param review_events: (boolean（可选）) 有评审事件触发回调钩子
        :param token: (string(可选)) 用以校验收到的负载;此token不会在包含在返回值中
        """
        return self.put("/projects/:id/hooks/:hook_id", id=id, hook_id=hook_id, **kwargs)

    def get_projects_hooks(self, id, hook_id, **kwargs):
        """

        查询项目单个回调钩子

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param hook_id: (integer) 回调钩子 id
        """
        return self.get("/projects/:id/hooks/:hook_id", id=id, hook_id=hook_id, **kwargs)

    def list_projects_hooks(self, id, **kwargs):
        """

        查询项目回调钩子列表

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/hooks", id=id, **kwargs)

    def delete_projects_hooks(self, id, hook_id, **kwargs):
        """

        删除项目回调钩子

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param hook_id: (integer) 回调钩子 id
        """
        return self.delete("/projects/:id/hooks/:hook_id", id=id, hook_id=hook_id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/merge_requests.md
    ##
    ##################################################################################

    def create_projects_merge_requests(self, id, **kwargs):
        """

        新增合并请求

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param source_branch: (string) 源分支
        :param target_branch: (string) 目标分支
        :param title: (string) 合并请求的标题
        :param assignee_id: (integer （可选）) 负责人 id
        :param description: (string  （可选）) 合并请求的描述
        :param target_project_id: (integer （可选）) 目标项目的 id
        :param labels: (string  （可选）) 合并请求的标签，多个请用英文逗号分隔
        :param reviewers: (string（可选）) 评审人id (只能是id。多个评审人用,隔开)
        :param necessary_reviewers: (string（可选）) 必要评审人id (只能是id。多个评审人用,隔开)
        :param approver_rule: (Integer（可选）) 评审人规则
        :param necessary_approver_rule: (Integer （可选）) 必要评审人规则
        """
        return self.post("/projects/:id/merge_requests", id=id, **kwargs)

    def update_projects_merge_request_merge(self, id, merge_request_id, **kwargs):
        """

        合并合并请求

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param merge_commit_message: (string  （可选）) 合并合并请求的描述消息
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id/merge",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def list_projects_merge_request_commits(self, id, merge_request_id, **kwargs):
        """

        获取合并请求中的提交

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (string) 合并请求的ID
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/commits",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def create_projects_merge_request_comments(self, id, merge_request_id, **kwargs):
        """

        添加合并请求的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param note: (integer) 评论内容
        """
        return self.post(
            "/projects/:id/merge_request/:merge_request_id/comments",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def list_projects_merge_request_comments(self, id, merge_request_id, **kwargs):
        """

        获取合并请求评论列表

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/comments",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_request(self, id, merge_request_id, **kwargs):
        """

        更新合并请求

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param target_branch: (string  （可选）) 目标分支
        :param assignee_id: (integer （可选）) 负责人 id
        :param title: (string  （可选）) 合并请求的标题
        :param state_event: (string  （可选）) 新的状态，可选值：（close|reopen\）
        :param description: (string  （可选）) 合并请求的描述
        :param labels: (string  （可选）) 合并请求的标签，多个请用英文逗号分隔
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def list_projects_merge_requests(self, id, **kwargs):
        """

        获取合并请求列表

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param iid: (integer（可选）) 项目里面的合并请求id编号
        :param state: (string （可选）) 合并请求状态，可选值：merged, opened 或 closed，不填写返回所有的合并请求
        :param order_by: (string （可选）) 排序字段， 允许按 created_at, updated_at 排序（默认 created_at）
        :param sort: (string （可选）) 排序方式， 允许 asc or desc（默认 desc）
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/merge_requests", id=id, **kwargs)

    def list_projects_merge_request_changes(self, id, merge_request_id, **kwargs):
        """

        查询合并请求的代码变更

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/changes",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def get_projects_merge_request(self, id, merge_request_id, **kwargs):
        """

        查询项目合并请求

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def get_projects_merge_request_subscribe(self, id, merge_request_id, **kwargs):
        """

        查询用户是否订阅请求合并

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 请求合并的 id
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/subscribe",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_request_subscribe(self, id, merge_request_id, **kwargs):
        """

        订阅请求合并

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id/subscribe",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_request_unsubscribe(self, id, merge_request_id, **kwargs):
        """

        取消订阅合并请求

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id/unsubscribe",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/comments.md
    ##
    ##################################################################################

    def create_projects_merge_requests_notes(self, id, merge_request_id, **kwargs):
        """

        创建合并请求的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param body: (string) 评论的内容
        :param _path: (string（可选）) 文件路径
        :param line: (string（可选）) 行号
        :param line_type: (string（可选）) 变更类型，可选old、new
        """
        return self.post(
            "/projects/:id/merge_requests/:merge_request_id/notes",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_requests_notes(self, id, merge_request_id, note_id, **kwargs):
        """

        编辑合并请求的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param note_id: (integer) 评论 id
        :param body: (string) 评论的内容
        """
        return self.put(
            "/projects/:id/merge_requests/:merge_request_id/notes/:note_id",
            id=id,
            merge_request_id=merge_request_id,
            note_id=note_id,
            **kwargs
        )

    def get_projects_merge_requests_notes(self, id, merge_request_id, note_id, **kwargs):
        """

        查询合并请求的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param note_id: (integer) 评论 id
        """
        return self.get(
            "/projects/:id/merge_requests/:merge_request_id/notes/:note_id",
            id=id,
            merge_request_id=merge_request_id,
            note_id=note_id,
            **kwargs
        )

    def get_projects_merge_requests_commits(self, id, merge_request_id, **kwargs):
        """

        查询合并请求的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param note_id: (integer) 评论 id
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/commits",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def list_projects_merge_requests_notes(self, id, merge_request_id, **kwargs):
        """

        查询合并请求的评论列表

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get(
            "/projects/:id/merge_requests/:merge_request_id/notes",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def get_projects_reviews_notes(self, id, review_id, note_id, **kwargs):
        """

        查询代码评审的单个评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审 id
        :param note_id: (integer) 评论 id
        """
        return self.get(
            "/projects/:id/reviews/:review_id/notes/:note_id",
            id=id,
            review_id=review_id,
            note_id=note_id,
            **kwargs
        )

    def list_projects_reviews_notes(self, id, review_id, **kwargs):
        """

        查询代码评审的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的 id
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get(
            "/projects/:id/reviews/:review_id/notes", id=id, review_id=review_id, **kwargs
        )

    def create_projects_reviews_notes(self, id, review_id, **kwargs):
        """

        创建代码评审的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审 id
        :param body: (string) 评论的内容
        :param _path: (string（可选）) 文件路径
        :param line: (string（可选）) 行号
        :param line_type: (string（可选）) 变更类型，可选old、new
        """
        return self.post(
            "/projects/:id/reviews/:review_id/notes", id=id, review_id=review_id, **kwargs
        )

    def update_projects_reviews_notes_note_id(self, id, review_id, **kwargs):
        """

        编辑代码评审的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审 id
        :param note_id: (integer) 评论 id
        :param body: (string) 评论的内容
        """
        return self.put(
            "/projects/:id/reviews/:review_id/notes/note_id", id=id, review_id=review_id, **kwargs
        )

    def create_projects_issues_notes(self, id, issue_id, **kwargs):
        """

        创建项目的缺陷评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷 id
        :param body: (string) 评论的内容
        """
        return self.post("/projects/:id/issues/:issue_id/notes", id=id, issue_id=issue_id, **kwargs)

    def update_projects_issues_notes(self, id, issue_id, note_id, **kwargs):
        """

        修改项目的缺陷评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷 id
        :param note_id: (integer) 评论 id
        :param body: (string) 评论 内容
        """
        return self.put(
            "/projects/:id/issues/:issue_id/notes/:note_id",
            id=id,
            issue_id=issue_id,
            note_id=note_id,
            **kwargs
        )

    def get_projects_issues_notes(self, id, issue_id, note_id, **kwargs):
        """

        查询单个缺陷的评论

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷 id
        :param note_id: (integer) 评论 id
        """
        return self.get(
            "/projects/:id/issues/:issue_id/notes/:note_id",
            id=id,
            issue_id=issue_id,
            note_id=note_id,
            **kwargs
        )

    def list_projects_issues_notes(self, id, issue_id, **kwargs):
        """

        获取缺陷评论列表

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param issue_id: (integer) 缺陷 id
        :param page: (integer) 分页（默认值：1）
        :param per_page: (integer) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/issues/:issue_id/notes", id=id, issue_id=issue_id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/cr/mr_review.md
    ##
    ##################################################################################

    def create_projects_merge_request_review_invite(self, id, merge_request_id, **kwargs):
        """

        邀请评审人

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param reviewer_id: (integer) 评审人的 id
        """
        return self.post(
            "/projects/:id/merge_request/:merge_request_id/review/invite",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def delete_projects_merge_request_review_dismissals(self, id, merge_request_id, **kwargs):
        """

        移除评审人

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        :param reviewer_id: (integer) 评审人 id
        """
        return self.delete(
            "/projects/:id/merge_request/:merge_request_id/review/dismissals",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def delete_projects_merge_request_review_cancel(self, id, merge_request_id, **kwargs):
        """

        取消评审

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        """
        return self.delete(
            "/projects/:id/merge_request/:merge_request_id/review/cancel",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def get_projects_merge_request_review(self, id, merge_request_id, **kwargs):
        """

        查询评审信息

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        """
        return self.get(
            "/projects/:id/merge_request/:merge_request_id/review",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_request_reviewer_summary(self, id, merge_request_id, **kwargs):
        """

        发表评审意见

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求的 id
        :param reviewer_event: (string) 评审人事件，可选：（comment | approve | require_change | deny）
        :param summary: (string) 评审信息摘要
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id/reviewer/summary",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    def update_projects_merge_request_review_reopen(self, id, merge_request_id, **kwargs):
        """

        重置评审状态

        :param id: (integer or string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param merge_request_id: (integer) 合并请求 id
        """
        return self.put(
            "/projects/:id/merge_request/:merge_request_id/review/reopen",
            id=id,
            merge_request_id=merge_request_id,
            **kwargs
        )

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/cr/code_review.md
    ##
    ##################################################################################

    def create_projects_review(self, id, **kwargs):
        """

        新建 Commit 评审

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param title: (string) 标题
        :param source_branch: (string) 源分支名（默认选择该分支最新的提交点）
        :param target_branch: (string) 目标分支名（默认选择该分支最新的提交点）
        :param description: (string （可选）) 描述
        :param source_commit: (string （可选）) 源提交点
        :param target_commit: (integer（可选）) 目标提交点
        :param target_project_id: (integer（可选）) 目标项目id
        :param reviewer_ids: (string（可选）) 评审人id (只能是id。多个评审人用,隔开)
        :param necessary_reviewer_ids: (string（可选）) 必要评审人id (只能是id。多个评审人用,隔开)
        :param approver_rule: (Integer（可选）) 评审人规则
        :param necessary_approver_rule: (Integer （可选）) 必要评审人规则
        """
        return self.post("/projects/:id/review", id=id, **kwargs)

    def list_projects_reviews(self, id, **kwargs):
        """

        获取项目中的Commit评审

        :param id: (integer 或string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param author_id: (integer（可选）) 作者的id
        :param state: (string（可选）) 代码状态，可选值：approving, approving change_required 或 closed，不填写返回所有的代码评审。
        :param order_by: (string（可选）) 排序字段， 允许按 created_at, updated_at 排序（默认 created_at）
        :param sort: (string（可选）) 排序方式， 允许 asc or desc（默认 desc）
        :param page: (integer（可选）) 分页（默认值：1）
        :param per_page: (integer（可选）) 默认页面大小（默认值： 20，最大值： 100）
        """
        return self.get("/projects/:id/reviews", id=id, **kwargs)

    def get_projects_review(self, id, review_id, **kwargs):
        """

        获取项目中某个具体的Commit评审

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        """
        return self.get("/projects/:id/review/:review_id", id=id, review_id=review_id, **kwargs)

    def create_projects_review_invite(self, id, review_id, **kwargs):
        """

        邀请评审人

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        :param reviewer_id: (Integer （可选）) 评审人id (只能是id。多个评审人用,隔开)
        :param necessary_reviewer_id: (Integer （可选）) 必要评审人id (只能是id。多个评审人用,隔开)
        """
        return self.post(
            "/projects/:id/review/:review_id/invite", id=id, review_id=review_id, **kwargs
        )

    def delete_projects_review_dismissals(self, id, review_id, **kwargs):
        """

        移除评审人

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        :param reviewer_id: (Integer) 评审人id
        """
        return self.delete(
            "/projects/:id/review/:review_id/dismissals", id=id, review_id=review_id, **kwargs
        )

    def update_projects_review_reviewer_summary(self, id, review_id, **kwargs):
        """

        发表评审意见

        :param id: (integer 或string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        :param reviewer_event: (Integer) 评审人事件，可选：（comment | approve | require_change | deny）
        :param summary: (string) 评审信息摘要
        """
        return self.put(
            "/projects/:id/review/:review_id/reviewer/summary", id=id, review_id=review_id, **kwargs
        )

    def update_projects_review_reopen(self, id, review_id, **kwargs):
        """

        重置代码评审状态

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        """
        return self.put(
            "/projects/:id/review/:review_id/reopen", id=id, review_id=review_id, **kwargs
        )

    def update_projects_review(self, id, review_id, **kwargs):
        """

        更新Commit评审

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param review_id: (integer) 代码评审的id
        :param title: (string) 代码评审的标题
        :param description: (string （可选）) 代码评审的描述
        """
        return self.put("/projects/:id/review/:review_id", id=id, review_id=review_id, **kwargs)

    ##################################################################################
    ##
    ## http://x.code.oa.com/Code/Git/blob/master/help/api/tloc_project.md
    ##
    ##################################################################################

    def get_projects_tloc_daily_count(self, id, **kwargs):
        """

        按天查询提交次数、提交用户数

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param begin_date: (yyyy-MM-dd'T'HH:mm:ssZ) 开始时间；例如2019-03-25T00:10:19+0000 或 2019-03-25T00:10:19+0800，时间参数必须转码，如“2019-03-25T00:10:19%2B0800”
        :param end_date: (yyyy-MM-dd'T'HH:mm:ssZ) 结束时间：例如2019-03-26T00:10:19+0000 或 2019-03-26T00:10:19+0800，时间参数必须转码，如“2019-03-26T00:10:19%2B0800”
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        :param timezone: (integer（可选）) 指定你的当前时区，默认是0时区，范围(-11,11)
        """
        return self.get("/projects/:id/tloc/daily/count", id=id, **kwargs)

    def get_projects_tloc_daily_diff(self, id, **kwargs):
        """

        按天查询提交代码行增量

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param begin_date: (yyyy-MM-dd'T'HH:mm:ssZ) 开始时间；例如2019-03-25T00:10:19+0000 或 2019-03-25T00:10:19+0800，时间参数必须转码，如“2019-03-25T00:10:19%2B0800”
        :param end_date: (yyyy-MM-dd'T'HH:mm:ssZ) 结束时间：例如2019-03-26T00:10:19+0000 或 2019-03-26T00:10:19+0800，时间参数必须转码，如“2019-03-26T00:10:19%2B0800”
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        :param timezone: (integer（可选）) 指定你的当前时区，默认是0时区，范围(-11,11)
        """
        return self.get("/projects/:id/tloc/daily/diff", id=id, **kwargs)

    def get_projects_tloc_user_diff(self, id, **kwargs):
        """

        查询各用户提交增量

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param begin_date: (yyyy-MM-dd'T'HH:mm:ssZ) 开始时间；例如2019-03-25T00:10:19+0000 或 2019-03-25T00:10:19+0800，时间参数必须转码，如“2019-03-25T00:10:19%2B0800”
        :param end_date: (yyyy-MM-dd'T'HH:mm:ssZ) 结束时间：例如2019-03-26T00:10:19+0000 或 2019-03-26T00:10:19+0800，时间参数必须转码，如“2019-03-26T00:10:19%2B0800”
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        """
        return self.get("/projects/:id/tloc/user/diff", id=id, **kwargs)

    def get_projects_tloc_lang_diff(self, id, **kwargs):
        """

        查询各语言提交增量

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param begin_date: (yyyy-MM-dd'T'HH:mm:ssZ) 开始时间；例如2019-03-25T00:10:19+0000 或 2019-03-25T00:10:19+0800，时间参数必须转码，如“2019-03-25T00:10:19%2B0800”
        :param end_date: (yyyy-MM-dd'T'HH:mm:ssZ) 结束时间：例如2019-03-26T00:10:19+0000 或 2019-03-26T00:10:19+0800，时间参数必须转码，如“2019-03-26T00:10:19%2B0800”
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        """
        return self.get("/projects/:id/tloc/lang/diff", id=id, **kwargs)

    def get_projects_tloc_commit_diff(self, id, **kwargs):
        """

        查询指定Commit点的提交增量

        :param id: (integer 或 string) id = 项目唯一标识或NAMESPACE___PATH/PROJECT___PATH
        :param sha: (string) commit hash值，必须是完整的SHA1，如：b1b9d46f1f6c55e8d157e41538b60bf7482ed436
        :param branch_name: (string（可选）) 指定分支名(默认全部分支)
        :param tag_name: (string（可选）) 指定Tag名(branch_name为空时生效，默认全部Tag)
        """
        return self.get("/projects/:id/tloc/commit/diff", id=id, **kwargs)



#!/usr/bin/env python3

"""
Author:         xiazeng
Create time:    2020/9/22
Description:    

"""
import inspect
import logging
import time
import datetime

import at
import at.core.uixml
import at.core.element
import at.core.uidevice
import at.core.exceptions
import case_repair_sdk

g_black_methods = ["to_main_ui", "_check_and_wait_status"]

logger = logging.getLogger()
g_test_id = None
g_seq_index = 0
g_run_mode = None
g_repair_sdk: case_repair_sdk.CaseRepairSdK = None
g_seq_info = {}
g_at_instance = None


class RepairConfig:
    ignore_check_failed = False  # 设置成True，在修复模式下，就不去检查修复的结果是否正确
    ignore_found_elem = False  # 设置成True，在修复模式下，如果能找到元素，就不做修复的逻辑


def add_block_methods(method_name):
    """不进行修复的函数"""
    if method_name not in g_black_methods:
        g_black_methods.append(method_name)


def set_run_mode(mode: str, test_id, at_instance, app_id) -> case_repair_sdk.CaseRepairSdK:
    """

    :param at_instance:
    :param mode: record/repair两种模式
    :return:
    """
    global g_run_mode, g_repair_sdk, g_at_instance, g_test_id
    if mode in ["record", "repair", "normal"]:
        g_run_mode = mode
        g_test_id = test_id
        g_repair_sdk = case_repair_sdk.CaseRepairSdK(test_id, app_id)
        g_at_instance = at_instance
        return g_repair_sdk
    else:
        raise RuntimeError("mode should be record or repair")


def set_ignore_check_failed(yes: bool):
    """
    设置忽略检查失败
    """
    RepairConfig.ignore_check_failed = yes


def is_record_mode():
    if g_run_mode == "record":
        return True
    return False


def is_normal_mode():
    if g_run_mode is None or g_run_mode == 'normal':
        return True
    return False


def is_repair_mode():
    if g_run_mode == "repair":
        return True
    return False


def ui_seq_start(seq_name, device_id=None):
    """
    用例开始运行时候上传
    :param seq_name:
    :param device_id 运行的手机ID，如果没有可以不传
    :return:
    """
    global g_seq_info
    if g_repair_sdk is None:
        return
    if device_id is None:
        device_id = int(time.time() % 1000000 * 1000)
    suffix = datetime.datetime.now().strftime("%H%M%S")
    # seq_id得保证唯一，为了人眼可快速定位，用test_id,device_id和时间拼一个id
    seq_id = f"{g_test_id}-{device_id}-{suffix}"
    g_seq_info = {
        "seq_name": seq_name,
        "seq_id": seq_id,
        "device_id": device_id,
        "device_desc": g_at_instance.adb.desc,
        "test_id": g_test_id,
        "is_finish": False,
        "mode": g_run_mode
    }
    g_repair_sdk.update_ui_seq(g_seq_info["seq_id"], g_seq_info)


def ui_seq_stop(success: bool, other_info=None):
    """
    用例结束运行时，上传，可以放到用例的teardown
    :param success:
    :param other_info:
    :return:
    """
    global g_seq_info, g_run_mode
    if g_repair_sdk is None:
        return
    g_run_mode = 'normal'
    g_seq_info["success"] = success
    g_seq_info["is_finish"] = True
    if other_info:
        if "case_name" in other_info:
            del other_info['case_name']
        if "seq_id" in other_info:
            del other_info['seq_id']
        g_seq_info.update(other_info)
    g_repair_sdk.update_ui_seq(g_seq_info["seq_id"], g_seq_info)
    g_seq_info = {}


def deco_do_action(f):
    """
    执行UI操作的装饰器
    :return:
    """

    def do_action(obj, *args, **kwargs):
        # 避免递归调用
        frames = inspect.stack()
        if len(frames) > 2:
            for frame in frames[1:]:
                method_name = frame[3]
                if method_name == "do_action":
                    return f(obj, *args, **kwargs)
                if method_name in g_black_methods:
                    logging.debug("ignore method:%s", method_name)
                    return f(obj, *args, **kwargs)

        if is_normal_mode():
            # 普通模式
            return f(obj, *args, **kwargs)
        elif not g_seq_info:
            # 录制或者回放模式，但是没有设置用例信息，一般是环境初始化的时候，跳过
            logger.error("case not start, do nothing, mode=%s", g_run_mode)
            return f(obj, *args, **kwargs)

        def get_xml_and_page(base_xml=None, retry=5):
            time.sleep(1)
            _xml_list = g_at_instance.java_driver.get_all_window_xmls()
            _xml_list = [g_at_instance.e.resguard.retrace_xml(_xml) for _xml in _xml_list]
            _page_name = g_at_instance.adb.get_current_activity()
            if _page_name is None:
                _page_name = "unknown"
            xml = None

            if len(_xml_list) > 1:
                # 删除无效的windows
                delete_index_list = []
                for i in range(len(_xml_list)):
                    ui_views = at.core.uixml.window_dump_parse_str(_xml_list[i])
                    if len(ui_views) <= 1:
                        logger.error("empty xml, delete it:%s, size:%s", _xml_list[i], len(_xml_list))
                        delete_index_list.append(i)
                delete_index_list.reverse()
                for i in delete_index_list:
                    del _xml_list[i]

            if retry == 0:
                return _xml_list[0], _page_name

            # 兼容逻辑，提高获取xml的稳定性
            if _page_name and 'com.tencent.mm' in _page_name:
                # 微信界面，但是拿到的xml确没有微信的信息
                if len(_xml_list) == 1 and 'com.tencent.mm' not in _xml_list[0]:
                    logger.error("is wechat activity, but xml not contains mm info, wait and try %s", retry)
                    return get_xml_and_page(base_xml, retry=retry - 1)

            for xml in _xml_list:
                if "android.widget.ProgressBar" in xml \
                        or 'com.tencent.mm:id/circle_progress_view' in xml \
                        :
                    logger.error("ProgressBar exists, wait and try %s", retry)
                    return get_xml_and_page(base_xml, retry=retry - 1)

            for xml in _xml_list:
                if "提醒" in xml and "多选" in xml:
                    # 微信会话弹框，会有多个窗口，默认用弹框里面的
                    if xml == base_xml:
                        logger.error("same as before, wait and try %s", retry)
                        return get_xml_and_page(base_xml, retry=retry - 1)
                    return xml, _page_name
            for xml in _xml_list:
                if "com.tencent.mm" in xml:
                    if xml == base_xml:
                        logger.error("same as before, wait and try %s", retry)
                        return get_xml_and_page(base_xml, retry=retry - 1)
                    return xml, _page_name
            for xml in _xml_list:
                if "com.android.systemui" not in xml:
                    if xml == base_xml:
                        logger.error("same as before, wait and try %s", retry)
                        return get_xml_and_page(base_xml, retry=retry - 1)
                    return xml, _page_name
            return _xml_list[0], _page_name

        def get_operation():
            operation = {
                "action": f.__name__,
                "widget": {

                },
                "x": -1,
                "y": -1,
                # 复杂的选项有问题
                "locator": {

                }
            }
            if isinstance(obj, at.core.element.Element):
                check_scroll = False
                if len(args) > 0 and args[0] == True:
                    check_scroll = True
                elif kwargs and kwargs.get("is_scroll"):
                    check_scroll = True

                max_scroll_num = 1
                width = g_at_instance.device.width()
                height = g_at_instance.device.height()
                if check_scroll:
                    max_scroll_num = 8
                    try:
                        list_view = at.core.element.Element().cls_name("android.widget.ListView").get_view()
                        logger.debug("max_scroll_num:%s, ui_view:%s", max_scroll_num, list_view)
                        if list_view:
                            width = list_view.width
                            height = list_view.height
                    except at.core.exceptions.UiNotFoundError:
                        logger.warning("max_scroll_num:%s, list view not found", max_scroll_num)

                from_points = [width / 2, height * 3 / 4]
                end_points = [width / 2, height * 1 / 4]
                steps = 12

                for i in range(max_scroll_num):
                    try:
                        ui_view = obj.get_view()
                        operation['locator'] = {
                            "type": "attr",
                            "value": ui_view.to_dict(),
                        }
                        operation["widget"] = {"x": ui_view.x1,
                                               "y": ui_view.y1,
                                               "w": ui_view.width,
                                               'h': ui_view.height
                                               }
                        return operation, True
                    except at.core.exceptions.UiNotFoundError:
                        logger.exception("need repair, because element not found:%s, remain scroll num:%s", obj,
                                         max_scroll_num - i - 1)
                        operation['locator'] = None
                        if i + 1 == max_scroll_num:
                            return operation, None
                    scroll_args = from_points + end_points
                    g_at_instance.device.scroll(*scroll_args, steps)

                # 滑动之后还是没有找到，恢复状态
                for i in range(max_scroll_num - 1):
                    scroll_args = end_points + from_points
                    g_at_instance.device.scroll(*scroll_args, steps)

                return operation, False

            elif isinstance(obj, at.core.uidevice.PyUiDevice):
                if f.__name__ == "click_on_point":
                    operation['x'], operation['y'] = args[0], args[1]
                else:
                    raise case_repair_sdk.RecordError("需要补充场景")
            else:
                raise case_repair_sdk.RecordError("不支持录制")

            return operation, True

        if is_record_mode():
            # 录制模式
            operation, found_it = get_operation()
            if not found_it:
                return f(obj, *args, **kwargs)
            before_xml, page_name = get_xml_and_page()
            before_img_base64 = g_at_instance.java_driver.request_at_device("screen", [1.0, 20, 200])

            f_ret = f(obj, *args, **kwargs)
            # 如果马上获取xml可能会出现中间状态
            time.sleep(2)
            after_xml, after_page_name = get_xml_and_page(before_xml)
            after_img_base64 = g_at_instance.java_driver.request_at_device("screen", [1.0, 20, 200])
            g_repair_sdk.upload_test_step(g_seq_info["seq_name"], page_name, g_seq_info["seq_id"], before_xml,
                                          after_xml, before_img_base64, after_img_base64, operation, after_page_name)
            return f_ret
        elif is_repair_mode():
            # 修复模式
            operation, found_it = get_operation()
            before_xml, page_name = get_xml_and_page()
            before_img_base64 = g_at_instance.java_driver.request_at_device("screen", [1.0, 20, 200])
            if found_it and RepairConfig.ignore_found_elem:
                return f(obj, *args, **kwargs)
            logger.debug("found elem:%s, operation:%s", found_it, operation)
            try:
                res, trace_id = g_repair_sdk.obtain_candidate(g_seq_info["seq_name"], page_name, g_seq_info["seq_id"],
                                                              before_xml, before_img_base64, operation)
            except:
                if found_it:
                    logger.exception("error, but element has found, ignore exception:")
                    return f(obj, *args, **kwargs)
                else:
                    raise

            repair_index = 0

            def try_repair(_candidate_item, _operation):
                # todo: 解析返回的数据
                logger.info("index:%s, candidate_item:%s, %s", repair_index, _candidate_item, _operation)
                if 'w' not in _candidate_item and 'rect' in _candidate_item:
                    x = int((_candidate_item['rect'][0] + _candidate_item['rect'][2]) / 2)
                    y = int((_candidate_item['rect'][1] + _candidate_item['rect'][3]) / 2)
                else:
                    x = _candidate_item['x'] + _candidate_item['w'] / 2
                    y = _candidate_item['y'] + _candidate_item['h'] / 2
                operation_action = _operation.get('action')
                if operation_action == "click" or operation_action == "click_on_point":
                    g_at_instance.device.click_on_point(x, y)
                elif operation_action == "long_click":
                    g_at_instance.device.long_click_on_point(x, y)
                elif operation_action == 'enter':
                    # 中午输入必须要用用java层输入，所以必须要在java层实现，比较复杂，暂时不做
                    g_at_instance.device.enter(args[0])
                    raise case_repair_sdk.RepairError("输入暂时不支持修复:%s" % _operation.get('action'))
                else:
                    raise case_repair_sdk.RepairError("unknown operation action:%s" % _operation.get('action'))

                time.sleep(2)
                return True

            if not res["need_repair"] or found_it:
                # 无需修复
                f_ret = f(obj, *args, **kwargs)
            elif not res['match_type']:
                # 当前状态完全不匹配
                raise case_repair_sdk.RepairError("empty match_type")
            else:
                # 进行修复
                candidate_item = res["candidate_item"]
                operation = res["operation"]
                repair_index += 1
                f_ret = try_repair(candidate_item, operation)

            # 校验修复是否正确
            while True:
                after_xml, page_name = get_xml_and_page(before_xml)
                before_xml = after_xml
                after_img_base64 = g_at_instance.java_driver.request_at_device("screen", [1.0, 20, 200])
                res = g_repair_sdk.check_repair(g_seq_info["seq_name"], page_name,
                                                g_seq_info["seq_id"], after_xml, after_img_base64, trace_id,
                                                operation)
                if res.get("need_repair"):
                    # 复杂case才会走到这个模式，如
                    # old path : page1 -> page2
                    # new path : page1 -> page' -> page2
                    f_ret = try_repair(res["candidate_item"], {'action': 'click'})
                elif res.get("is_repaired"):
                    break
                else:
                    if RepairConfig.ignore_check_failed:
                        logger.error("验证失败，但是配置忽略了验证失败，继续往下走")
                        break
                    else:
                        raise case_repair_sdk.CheckError("验证修复后的结果失败")
            return f_ret
        else:
            raise case_repair_sdk.RepairError("不能解析的运行模式")

    return do_action



#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Author:         yopofeng
Filename:       utils.py
Create time:    2021/9/24 21:21
Description:

"""
import time
import typing
from typing import overload
import types
import platform
import re
import sys
import inspect
import threading
import logging
import os
import asyncio
import functools
import concurrent.futures
import json
import socket

try:
    import urllib
    urlencode = urllib.urlencode
except AttributeError:
    from urllib.parse import urlencode
except ImportError:
    try:
        from urllib3.request import urlencode
    except ModuleNotFoundError:  # 2.0.2 后改了
        raise RuntimeError("maybe you should install urllib3<2.0")

logger = logging.getLogger("minium")


def timeout(duration, interval=1):
    """
    重试超时装饰器,在超时之前会每隔{interval}秒重试一次
    注意：被修饰的函数必须要有非空返回,这是重试终止的条件！！！
    :param duration: seconds
    :return:
    """

    def spin_until_true(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            timeout = time.time() + duration
            execed = False
            r = None
            while not (r or timeout < time.time() and execed):
                r = func(*args, **kwargs)
                execed = True
                if r or timeout < time.time():
                    return r
                time.sleep(interval)
            return r

        return wrapper

    return spin_until_true


def retry(cnt, expected_exception=None):
    """
    重试固定次数装饰器, 被修饰函数没有raise error则直接返回, 有则最多执行${cnt}次
    :cnt: 重试次数，函数最多被执行 ${cnt} 次
    :expected_exception: 命中预期的错误(isinstance(e, expected_exception))才重试, None为所有错误
    """

    def try_until_no_error(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _cnt = 0
            while _cnt < cnt:
                try:
                    _cnt += 1
                    return func(*args, **kwargs)
                except:
                    if _cnt >= cnt:
                        raise
                    e = sys.exc_info()[1]
                    if expected_exception is None or isinstance(e, expected_exception):
                        logger.warning(
                            f'{f"第 {_cnt} 次" if _cnt < cnt else "最后一次"}重新运行{func.__name__}'
                        )
                        logger.info(f'原因: {e.__class__.__name__}{str(e.args)}')
                        continue
                    raise

        return wrapper

    return try_until_no_error


CatchableType = typing.Union[types.FunctionType, types.MethodType]


@overload
def catch(wrapped: CatchableType, *args: typing.Tuple[Exception, ...]) -> CatchableType:
    ...


@overload
def catch(*args: typing.Tuple[Exception, ...]) -> CatchableType:
    ...


def catch(*args):
    """
    抓获指定/所有exception
    :wrapped: 被修饰的函数, 如果为空则作为修饰器使用
    :expected_exception: 指定/所有exception
    :return: Exception/None

    etc.
    @catch
    def catchAllException():
        raise Exception("test")

    @catch(ValueError, RuntimeError)
    def catchValueAndRuntimeError():
        raise ValueError("test")

    def raiseValueError():
        raise ValueError("test")
    catch(raiseValueError)() -> ValueError("test")
    catch(raiseValueError, ValueError)() -> ValueError("test")
    catch(raiseValueError, ValueError, RuntimeError)() -> ValueError("test")
    catch(ValueError, RuntimeError)(raiseValueError)() -> ValueError("test")
    catch(raiseValueError, RuntimeError)() -> raise ValueError("test")
    """
    wrapped = None
    expected_exception = None
    if len(args) == 0:
        expected_exception = None
    elif len(args) >= 1:
        if inspect.isfunction(args[0]) or inspect.ismethod(args[0]):
            wrapped = args[0]
            expected_exception = args[1:] or None
        else:
            expected_exception = args

    def try_catch(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except:
                e = sys.exc_info()[1]
                if expected_exception is None or isinstance(e, expected_exception):
                    logger.error("catch error: %s", e)
                    return e
                raise

        return wrapper

    if wrapped:
        return try_catch(wrapped)
    # 修饰器用法
    return try_catch


class WaitTimeoutError(TimeoutError):
    pass


TimeoutErrors = (TimeoutError, WaitTimeoutError, asyncio.exceptions.TimeoutError)


class WaitThread(threading.Thread):
    def __init__(
        self,
        group=None,
        target=None,
        name=None,
        args=(),
        kwargs=None,
        daemon=None,
        semaphore=None,
    ):
        super().__init__(group, target, name, args, kwargs, daemon=daemon)
        self._result = None
        self._exception = None
        self._semaphore: threading.Semaphore = semaphore

    def run(self):
        try:
            if self._target:
                self._result = self._target(*self._args, **self._kwargs)
        except:
            self._exception = sys.exc_info()[1]
        finally:
            # Avoid a refcycle if the thread is running a function with
            # an argument that has a member that points to the thread.
            del self._target, self._args, self._kwargs
            if self._semaphore:
                self._semaphore.release()

    def get_result(self, timeout=None, block=True):
        if block:
            self.join(timeout=timeout)
        if self._exception:
            raise self._exception
        if block and self.is_alive():
            raise WaitTimeoutError("wait [%s] seconds timeout" % timeout)
        return self._result


def wait(timeout, default=None):
    """
    等待修饰器
    等待timeout时间, 如果函数没有返回则返回default
    """

    def spin_until_true(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            t = WaitThread(target=func, args=args, kwargs=kwargs)
            t.setDaemon(True)
            t.start()
            try:
                return t.get_result(timeout)
            except WaitTimeoutError:
                logger.error("wait %s %ss timeout" % (func.__name__, timeout))
                return default

        return wrapper

    return spin_until_true


_platform = platform.platform()
isWindows = "Windows" in _platform
isMacOS = "Darwin" in _platform or "macOS" in _platform


class Version(object):
    def __init__(self, version: str) -> None:
        m = re.match(r"^([0-9\.]+|latest|dev)", version or "")
        if not m:
            raise ValueError(f"{version} format not collect")
        self.version = "latest" if m.group(1) in ("latest", "dev") else m.group(1)

    def __str__(self) -> str:
        return self.version

    def __comp_version(self, a: str, b: str) -> int:
        """
        description: 对比基础库版本
        param {*} self
        param {str} a
        param {str} b
        return {int} 1 if a > b, 0 if a == b ,-1 if a < b
        """
        latest = ("latest", "dev")  # latest, dev版本看作是最大的版本号
        if a in latest:
            return 0 if b in latest else 1
        if b in latest:
            return 0 if a in latest else 1
        i = 0
        a = a.split(".")
        b = b.split(".")
        while i < len(a) and i < len(b):
            if int(a[i]) > int(b[i]):
                return 1
            elif int(a[i]) < int(b[i]):
                return -1
            i += 1
        return 0

    def __lt__(self, version):
        if isinstance(version, str):
            version = Version(version)
        if self.__comp_version(self.version, version.version) == -1:
            return True
        return False

    def __gt__(self, version):
        if isinstance(version, str):
            version = Version(version)
        if self.__comp_version(self.version, version.version) == -1:
            return False
        return True

    def __le__(self, version):
        if isinstance(version, str):
            version = Version(version)
        if self.__comp_version(self.version, version.version) != 1:
            return True
        return False

    def __ge__(self, version):
        if isinstance(version, str):
            version = Version(version)
        if self.__comp_version(self.version, version.version) != -1:
            return True
        return False

    def __eq__(self, version):
        if isinstance(version, str):
            version = Version(version)
        return self.version == version.version


def add_path_to_env(path: str):
    """
    把路径添加到PATH环境变量中
    """
    SPLIT = ":" if isMacOS else ";"
    _path = path if os.path.isdir(path) else os.path.dirname(path)
    env_paths = (os.environ["PATH"] or "").split(SPLIT)
    if _path not in env_paths:
        env_paths.append(_path)
    os.environ["PATH"] = SPLIT.join(env_paths)


class ProcessSafeEventLoop(object):
    def __init__(self) -> None:
        self.pid = os.getpid()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.run_loop()

    def run_loop(self):
        threading.Thread(target=self.loop.run_forever, daemon=True).start()

    def stop_loop(self):
        self.loop.stop()

    def run_coroutine(self, coro):
        """Submit a coroutine object to a given event loop.

        Return a concurrent.futures.Future to access the result.
        """
        if os.getpid() != self.pid:
            self.pid = os.getpid()
            self.loop = asyncio.new_event_loop()
            self.run_loop()
        return asyncio.run_coroutine_threadsafe(coro, loop=self.loop)

    def is_running(self):
        return self.loop.is_running()

    def __getattr__(self, name):
        return getattr(self.loop, name)


async def _cancel_and_wait(fut, loop):
    """Cancel the *fut* future or task and wait until it completes."""

    waiter = loop.create_future()
    cb = functools.partial(_release_waiter, waiter)
    fut.add_done_callback(cb)

    try:
        fut.cancel()
        # We cannot wait on *fut* directly to make
        # sure _cancel_and_wait itself is reliably cancellable.
        await waiter
    finally:
        fut.remove_done_callback(cb)


def _release_waiter(waiter, *args):
    if not waiter.done():
        waiter.set_result(None)


async def async_wait(
    fut, timeout, loop: ProcessSafeEventLoop or asyncio.AbstractEventLoop = None
):
    """
    reference asyncio.wait_for
    wait fut done, when timeout, raise
    """
    if loop is None:
        loop = asyncio.get_running_loop()
    elif isinstance(loop, ProcessSafeEventLoop):
        loop = loop.loop
    else:
        loop = loop
    waiter = loop.create_future()
    cb = functools.partial(_release_waiter, waiter)
    fut = asyncio.ensure_future(fut, loop=loop)
    fut.add_done_callback(cb)
    timeout_handle = loop.call_later(timeout, _release_waiter, waiter)

    try:
        # wait until the future completes or the timeout WaitTimeoutError
        try:
            await waiter
        except asyncio.exceptions.CancelledError:
            if fut.done():
                return fut.result()
            else:
                fut.remove_done_callback(cb)
                # We must ensure that the task is not running
                # after wait_for() returns.
                # See https://bugs.python.org/issue32751
                await _cancel_and_wait(fut, loop=loop)
                raise

        if fut.done():
            return fut.result()
        else:
            fut.remove_done_callback(cb)
            # We must ensure that the task is not running
            # after wait_for() returns.
            # See https://bugs.python.org/issue32751
            await _cancel_and_wait(fut, loop=loop)
            # In case task cancellation failed with some
            # exception, we should re-raise it
            # See https://bugs.python.org/issue40607
            try:
                fut.result()
            except asyncio.exceptions.CancelledError as exc:
                raise WaitTimeoutError() from exc
            else:
                raise WaitTimeoutError()
    finally:
        timeout_handle.cancel()


class AsyncCondition(asyncio.Condition):
    def __init__(
        self,
        lock: asyncio.Lock = None,
        *,
        loop: asyncio.AbstractEventLoop or ProcessSafeEventLoop = None,
    ) -> None:
        if isinstance(loop, ProcessSafeEventLoop):
            loop = loop.loop
        if sys.version_info < (3, 10):
            super().__init__(lock, loop=loop)
        else:  # loop参数将在3.10废除
            asyncio.set_event_loop(loop)
            super().__init__(lock)

    async def wait(self, timeout=None):
        loop = self._loop
        coro = super().wait()
        if timeout is None:
            return await coro
        try:
            return await async_wait(coro, timeout=timeout, loop=loop)
        except WaitTimeoutError:
            return False


Future = typing.Union[asyncio.futures.Future, concurrent.futures.Future]
EventLoop = typing.Union[ProcessSafeEventLoop, asyncio.BaseEventLoop]


def get_result(
    fut: Future,
    timeout=None,
    default=None,
):
    # asyncio.futures.Future.result 去掉了timeout参数
    if isinstance(fut, concurrent.futures.Future):
        try:
            return fut.result(timeout)
        except concurrent.futures.TimeoutError as ext:
            if default is not None:
                return default
            raise WaitTimeoutError() from ext
    loop = fut.get_loop()
    try:
        return asyncio.run_coroutine_threadsafe(
            async_wait(fut, timeout=timeout, loop=loop), loop=loop
        ).result()
    except WaitTimeoutError as ext:
        if default is not None:
            return default
        raise WaitTimeoutError() from ext


class Object(dict):
    def __init__(self, __map=None, **kwargs):
        if __map:
            kwargs.update(__map)
        extend = {}
        for k, v in kwargs.items():
            if hasattr(dict, k):  # dict本来的属性不可覆盖
                extend[k] = v
                continue
            setattr(self, k, v)
        super(Object, self).__init__(self.__dict__, **extend)

    def __getattr__(self, __k):
        try:
            return self[__k]
        except KeyError:
            return None

    def __setattr__(self, __k, __v):
        if isinstance(__v, dict):
            __v = Object(__v)
        if isinstance(__v, list):
            for index, v in enumerate(__v):
                if isinstance(v, dict):
                    __v[index] = Object(v)
        if hasattr(self.__class__, __k):
            super(Object, self).__setattr__(__k, __v)
        else:
            self[__k] = __v

    @classmethod
    def parse_from_file(cls, file_path):
        if not os.path.isfile(file_path):
            raise RuntimeError(f"{file_path} not exists")
        with open(file_path, "r", encoding="utf8") as fp:
            return cls(json.load(fp))
        
def pick_unuse_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("localhost", 0))
    addr, port = s.getsockname()
    s.close()
    return port



# -*- coding: utf-8 -*-
"""
@author: 'xiazeng'
@created: 2016/11/23
"""
# import unittest
import json
import typing
import os.path
import inspect
import time
from functools import wraps
from .logcolor import *
import logging.handlers
import traceback
import datetime
import re
from minium.miniprogram.wx_minium import WXMinium
from .miniconfig import MiniConfig, get_log_level
from minium.framework.libs import unittest

logger = logging.getLogger("minium")
logger.propagate = False
WORKSPACE_DIR = os.path.abspath(os.getcwd())
# print("当前工作路径: %s" % WORKSPACE_DIR)
# LOG_FORMATTER = "%(levelname)-5.5s %(asctime)s %(filename)-10s %(funcName)-15s %(lineno)-3d %(message)s"
FILENAME_LOGGER = "loader{0}.log".format(datetime.datetime.now().strftime("%Y%m%d"))

g_case_log_handler = None

g_from_command = False

__ALL__ = ["AssertBase", "exit_when_error"]


def exit_when_error(test_item):
    setattr(test_item, "__stop_when_error", True)
    return test_item

AssertBase = None

def hook_wrapper(src):
    @wraps(src)
    def hook_assert(self: AssertBase, *args, **kwargs):
        if not self._hook_assert:
            return src(self, *args, **kwargs)
        called_frame = inspect.getouterframes(inspect.currentframe(), 2)
        called_name = called_frame[1][3]
        signature = inspect.signature(src)
        try:
            msg_index = list(signature.parameters.keys()).index("msg")
        except ValueError:
            msg_index = -1
        ret = True
        print_msg = None
        try:
            return src(self, *args, **kwargs)
        except AssertionError as e:
            self._has_assert_error = True
            ret = False
            print_msg = str(e)
            raise
        finally:
            # todo:
            #  1. msg needed
            if called_name and not called_name.startswith("assert"):
                if msg_index < len(args) and msg_index > -1:
                    name = args[msg_index]
                else:
                    name = kwargs.get("msg")
                if name:
                    name = re.sub(r"\W|^(?=\d)", "_", name)
                self._add_assert_info(name or src.__name__, ret, print_msg)
    return hook_assert

class HookAssert(type):
    def __new__(cls, name, base, attr_dict):
        def search_parent(bases):
            # print("search", bases)
            for b in bases:
                for k in b.__dict__.keys():
                    if k not in attr_dict and k.startswith("assert") and callable(b.__dict__[k]):
                        attr_dict[k] = b.__dict__[k]  # 算直接继承?
                if not isinstance(b, HookAssert):  # 往上搜索
                    search_parent(b.__bases__)
        search_parent(base)
        for k in attr_dict:
            if k.startswith("assert") and callable(attr_dict[k]) and not hasattr(attr_dict[k], "__wrapped__"):
                attr_dict[k] = hook_wrapper(attr_dict[k])
        return super().__new__(cls, name, base, attr_dict)

class AssertBase(unittest.TestCase, metaclass=HookAssert):
    DEVICE_INFO = {}
    CONFIG = None

    test_config: MiniConfig = None

    _hook_assert = False

    def _startTest(self):
        # 初始化配置
        self._setup_config()
        # 初始化hook assert
        self._setup_assert()
        # 初始化日志
        self._setup_log()
        # 初始化结果数据
        self._setup_results()

    def _endTest(self):
        self._teardown_handle()
        self._teardown_collect()
        self._teardown_result()
        self._teardown_log()

    # ===============================================================================================
    # Setup
    # ===============================================================================================
    def _miniSetUp(self):
        # logger.info("_miniSetUp")
        super(AssertBase, self)._miniSetUp()
        if not getattr(
            self, "__skip_mini_init__", None
        ):  # 初始化output需要放在result.starttest中，考虑到不使用miniresult的情况，在case.setup中需要检查一下配置初始化
            self._startTest()

    @classmethod
    def setUpConfig(cls, default=False):
        if cls.CONFIG is None:
            default_config_filename = os.path.join(WORKSPACE_DIR, "config.json")
            if os.path.exists(default_config_filename) and not default:
                logger.info("read config from workspace dir: [%s]" % WORKSPACE_DIR)
                cls.CONFIG = MiniConfig.from_file(default_config_filename)
            else:
                logger.warning("loading default config")
                cls.CONFIG = MiniConfig()
        if cls.CONFIG.outputs is None:
            outputs = os.path.join(os.getcwd(), "outputs")
            if not os.path.exists(outputs):
                os.makedirs(outputs)
            cls.CONFIG.outputs = outputs
        if cls.CONFIG.create_time is None:
            cls.CONFIG.create_time = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        global g_case_log_handler
        log_level = get_log_level(cls.CONFIG.debug_mode)
        g_console_handler.setLevel(log_level)

        if g_case_log_handler is None:  # 格式化 -> 固化到文件
            log_path = os.path.join(cls.CONFIG.outputs, FILENAME_LOGGER)
            case_log_handler = logging.handlers.RotatingFileHandler(
                log_path, maxBytes=1024 * 1024, encoding="utf-8"
            )
            case_log_handler.addFilter(nofilter)
            formatter = logging.Formatter(LOG_FORMATTER)
            case_log_handler.setFormatter(formatter)
            case_log_handler.setLevel(logging.DEBUG)
            logger.addHandler(case_log_handler)
            logger.setLevel(log_level)
            g_case_log_handler = case_log_handler

        # 每次加载配置都要重新设置日志等级
        logger.setLevel(log_level)

    @classmethod
    def _miniClassSetUp(cls):
        cls.setUpConfig()

    @classmethod
    def setUpClass(cls) -> None:
        if not getattr(cls, "__is_minitest_suite__", False):
            _miniClassSetUp = getattr(cls, "_miniClassSetUp", None)
            if _miniClassSetUp:
                _miniClassSetUp()
        super(AssertBase, cls).setUpClass()

    def _setup_config(self):
        """
        测试结果存放目录
        1.0 版本存放目录
        ${self.test_config.outputs}/${self._testMethodName}/${datetime.datetime.now().strftime("%Y%m%d%H%M%S")}
        2.0 版本存放目录
        ${self.test_config.outputs}/${self.test_config.create_time}/${self._testMethodName}/
        ${datetime.datetime.now().strftime("%Y%m%d%H%M%S")}
        :return:
        """
        self.test_config = MiniConfig(self.CONFIG)
        dt = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
        self.test_config.case_output = os.path.join(
            self.test_config.outputs,
            self.test_config.create_time,
            self._testMethodName,
            dt,
        )
        if not os.path.exists(self.test_config.case_output):
            os.makedirs(self.test_config.case_output)

    def _setup_assert(self):
        self._hook_assert = True
        self._has_assert_error = False
        self.assert_list = []
        self.__assert_index = 0

    def _setup_log(self):
        # 保存日志到outputs
        # case_log_filename = self.wrap_filename("{0}.log".format(self._testMethodName))
        case_log_filename = self.wrap_filename("run.log")
        self._log_filename = os.path.basename(case_log_filename)
        case_log_handler = logging.handlers.RotatingFileHandler(
            case_log_filename, maxBytes=1024 * 1024, encoding="utf-8"
        )
        case_log_handler.addFilter(nofilter)  # 过滤颜色
        case_log_handler.addFilter(kwfilter)  # 过滤无关的关键字
        case_log_handler.addFilter(filefilter)  # 过滤无关的文件log
        formatter = logging.Formatter(LOG_FORMATTER)
        case_log_handler.setFormatter(formatter)
        self._case_log_handler = case_log_handler
        logger.addHandler(self._case_log_handler)
        # logger.info(self.CONFIG)

    def _setup_results(self):
        self.setup_time = time.time()
        self.screen_info = []
        self.step_info = []  # 录制回放用
        self.check_list = []
        desc = self._testMethodDoc
        if desc is not None:
            lines = desc.split("\n")
            lines = [l.strip() for l in lines if l]
            desc = "\n".join(lines)
        module_filename = inspect.getsourcefile(self.__class__)
        self._test_filename = module_filename
        module_name = None
        if module_filename:
            module_filename = os.sep.join(module_filename.split(".")[:-1])
            tokens = module_filename.split(os.sep)
            if len(tokens) == 1:
                tokens = module_filename.split("/")
            root_package_name = os.path.abspath(__file__).split(os.sep)[-3]
            if root_package_name in tokens:
                index = tokens.index(root_package_name)
                module_name = ".".join(tokens[index + 2:])
            else:
                # 支持启动类不在MMWebTest子目录下
                module_name = ".".join(tokens)
        packages = self.__module__.split(".")
        self.results = {
            "case_name": self._testMethodName,
            "run_time": str(datetime.datetime.now().strftime("%Y%m%d %H:%M:%S")),
            "test_type": self.__class__.__name__,
            "case_doc": desc,
            "success": False,
            "failures": "",
            "errors": "",
            "start_timestamp": self.setup_time,
            "is_failure": False,
            "is_error": False,
            "module": module_name,
            "package": ".".join(packages[:-2]) if len(packages) > 2 else packages[0],
            "failed_line_num": -1,
            "device": AssertBase.DEVICE_INFO,
            "log_filename": self._log_filename,
            "error_type": "",
            "error_value": "",
            "error_stages": [],
            "screen_info": self.screen_info,
            "step_info": self.step_info,
            "check_list": self.check_list,
            "assert_list": self.assert_list,
        }

    # ===============================================================================================
    # Teardown
    # ===============================================================================================
    def _teardown_handle(self):
        """
        teardown收集数据前进行的操作
        """
        pass

    def _get_error_info(self) -> typing.Tuple[typing.Any, list, str]:
        if not self._outcome:
            logger.error("self._outcome is None, maybe a class error")
            sys_info = getattr(
                self.__class__, "__class_setup_failure_why__", None
            )
            error_stages = getattr(
                self.__class__, "__class_setup_failure_stages__", []
            )
            stage = error_stages[-1] if error_stages and len(error_stages) > 0 else ""
            return sys_info, error_stages, stage # 看看是不是class失败了
        sys_info = None
        error_stages = getattr(self._outcome, "error_stages", None)
        stage = ""
        if error_stages is not None:  # minium.unittest.outcome
            stage_errors = getattr(self._outcome, "stage_errors", {})
            if "testMethod" in error_stages and "testMethod" in stage_errors:  # case内容部分有错误
                sys_info = stage_errors["testMethod"][-1]  # 优先处理case错误
                stage = "testMethod"
            else:
                for i in range(len(error_stages)-1, -1, -1):
                    if error_stages[i] in stage_errors:
                        stage = error_stages[i]
                        sys_info = stage_errors[stage][-1]
                        break
            return sys_info, error_stages, stage
        else:
            length = len(self._outcome.errors)
            if length == 1:
                sys_info = self._outcome.errors[-1][-1]
            elif length > 1:
                while length:
                    length -= 1
                    sys_info = self._outcome.errors[length][-1]
                    if sys_info is not None:  # step success
                        return sys_info, [], stage
            return sys_info, [], stage
        
    def _get_source(self, stage) -> dict:
        if not stage:
            stage = "testMethod"
        source_line = [[], 0]
        test_method = getattr(self, self._testMethodName if stage == "testMethod" else stage, None)
        if test_method:
            source_line = inspect.getsourcelines(test_method)
        return {"code": source_line[0], "start": source_line[1]}

    def _get_stack_info(self, stage, sys_info):
        if not sys_info:
            return
        failed_line_num = -1
        isDdt = False
        if stage == "testMethod":
            test_method = getattr(self, self._testMethodName, None)
            if test_method and getattr(test_method, "__wrapped__", None):
                # testMethod.__wrapped__存在, 为ddt生成的方法
                isDdt = True
                logger.warn(f"{self._testMethodName} isDdt case")
        else:
            test_method = getattr(self, stage, None)
        e_type, e_value, e_traceback = sys_info
        stack_lines = traceback.format_exception(e_type, e_value, e_traceback)
        logger.exception(f"{test_method}", exc_info=sys_info)
        if len(stack_lines) > 2:
            r = re.compile(r"File [\'\"](.*)[\'\"], line (\d+), in (\S+)")
            for stack_line in stack_lines[1:]:
                m = r.search(stack_line)
                if m and stage == "testMethod":
                    if (
                        m.group(1) == self._test_filename
                        and (self._testMethodName.startswith(m.group(3)) if isDdt else m.group(3) == self._testMethodName)
                    ):
                        failed_line_num = int(m.group(2))
                elif m and m.group(3) == stage:
                    failed_line_num = int(m.group(2))
        return {"failed_line_num": failed_line_num, "stack_lines": stack_lines, "error_type": e_type.__name__, "error_value": str(e_value.args[0]) if len(e_value.args) > 0 else ""}

    def _teardown_collect(self):
        """
        收集results数据
        """
        self.results["stop_timestamp"] = time.time()
        sys_info, error_stages, stage = self._get_error_info()

        if not self.test_config.only_native:
            self.results["appId"] = self.appId
            self.results["appName"] = self.appName
        else:
            logger.warning("Only native mode, skip get APP id and APP nickname")
        self.results["source"] = self._get_source(stage)
        if sys_info:
            stack_info = self._get_stack_info(stage, sys_info)
            self.results["failed_line_num"] = stack_info["failed_line_num"]
            if self._has_assert_error:
                self.results["failures"] = "".join(stack_info["stack_lines"])
                self.results["is_failure"] = True
            else:
                self.results["errors"] = "".join(stack_info["stack_lines"])
                self.results["is_error"] = True
            self.results["error_type"] = stack_info["error_type"]
            self.results["error_value"] = stack_info["error_value"]
            self.results["error_stages"] = error_stages
        else:
            self.results["success"] = True

    def _teardown_log(self):
        """
        落地log
        """
        logger.removeHandler(self._case_log_handler)
        self._case_log_handler.close()

    def _teardown_result(self):
        """
        落地结果数据
        """
        # filename = self.wrap_filename("{0}.json".format(self._testMethodName))
        filename = self.wrap_filename("result.json")
        self.results["filename"] = os.path.basename(filename)
        with open(filename, "w", encoding="UTF-8") as f:
            json.dump(self.results, f, indent=4)

    def _miniTearDown(self):
        # logger.info("_miniTearDown")
        if not getattr(self, "__skip_mini_init__", None):
            self._endTest()

    # ===============================================================================================
    # assert 拦截， 在进行assert开始和拦截之前做一些数据操作
    # ===============================================================================================
    def _add_assert_info(self, name, ret, reason=None):
        self.__assert_index += 1
        try:
            hook_result = self.hook_assert(name, ret, reason)
        except Exception as e:
            logger.exception(e)
            hook_result = None
        self.assert_list.append(
            {
                "name": name,
                "ret": ret,
                "msg": reason,
                **(hook_result if isinstance(hook_result, dict) else {}),
            }
        )

    @property
    def screen_dir(self):
        screen_dir = self.wrap_filename("images")
        if not os.path.exists(screen_dir):
            os.makedirs(screen_dir)
        return screen_dir

    def add_screen(self, name, path, url):
        self.screen_info.append(
            {
                "name": name,
                "url": url,
                "path": self.get_relative_path(path),
                "ts": int(time.time()),
                "datetime": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            }
        )

    def wrap_filename(self, filename):
        return os.path.abspath(os.path.join(self.test_config.case_output, filename))

    def get_relative_path(self, path):
        output = self.test_config.case_output
        if not os.path.isabs(output):
            output = os.path.abspath(output)
        if not output.endswith(os.path.sep):
            output = output + os.path.sep
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        if not path.startswith(output):
            logger.error("%s not in outputs: %s" % (path, output))
        else:
            path = path[len(output) :]
        return path

    # ===============================================================================================
    # hooks
    # ===============================================================================================
    def hook_assert(self, name, ret, reason=None) -> dict or None:
        """
        在assert校验的时候被调用
        :return: dict
        """
        pass



#!/usr/bin/env python3
# Created by xiazeng on 2019-05-22
import os
import json
import logging
import yaml
import copy

logger = logging.getLogger("minium")


class Path(str):
    pass


default_config = {
    "debug": False,                             # debug模式
    "base_dir": os.path.abspath(os.getcwd()),   # 如果没有配置, 默认工作目录, 如果通过文件生成则为文件所在目录
    "platform": "ide",                          # 平台: ide, android, ios
    "app": "wx",                                # 承载的app: wx
    "debug_mode": "debug",                      # 日志级别
    "close_ide": False,                         # 是否关闭IDE
    # "assert_capture": True,                   # 断言时是否截图, 不再对外. 仍可使用, 如只想setup&teardown时截图, 可以配置auto_capture: auto & assert_capture: false
    # "framework_capture": True,                # setup, teardown时自动截图
    # "error_capture": True,                    # case报错时自动截图
    "auto_capture": "auto",                     # auto: setup, teardown, assert时自动截图. error: 只有case报错时自动截图. assert: 断言时自动截图. False/"": 不截图
    "auto_relaunch": True,                      # case开始时是否回到主页
    "device_desire": {},                        # 真机调试配置
    "account_info": {},                         # 账号配置
    "report_usage": True,                       # 是否需要报告
    "remote_connect_timeout": 180,              # 真机调试中小程序在真机上打开的等待时间
    "request_timeout": 60,                      # 自动化控制指令请求超时时间
    "use_push": True,                           # 真机调试中是否使用推送形式打开小程序, false则需要扫调试二维码
    "full_reset": False,                        # 每个测试class结束，是否释放调试链接
    "outputs": None,                            # 测试产物输出路径
    "enable_app_log": False,                    # 记录小程序日志
    "enable_network_panel": False,              # 记录小程序网络请求
    "project_path": None,                       # 小程序项目路径
    "dev_tool_path": None,                      # 开发者工具命令行工具路径
    "test_port": 9420,                          # 小程序自动化测试调试端口
    "mock_native_modal": {},                    # 仅在IDE生效, mock所有会有原生弹窗的接口
    "mock_request": [],                         # mock request接口，item结构为{rule: {}, success or fail}, 同app.mock_request参数
    "auto_authorize": False,                    # 自动处理授权弹窗
    "audits": None,                             # 开启体验评分，仅在IDE生效, None为不改变
    "teardown_snapshot": False,                 # teardown快照，纪录page.data & page.wxml
    "mock_images_dir": Path(""),                # 需要进行"上传"操作的图片放置的目录
    "mock_images": {},                          # 需要进行"上传"操作的图片, key为图片名(标识用), value为base64格式的图片数据
    "need_perf": False,                         # 需要性能数据
    "appid": None,                              # 小程序appid
}


def get_log_level(debug_mode):
    return {
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "warn": logging.WARNING,
        "error": logging.ERROR,
    }.get(debug_mode, logging.INFO)


class MiniConfig(dict):
    def __init__(self, from_dict=None):
        default = copy.deepcopy(default_config)
        if isinstance(from_dict, dict):
            for k, v in from_dict.items():
                if k in default_config and isinstance(default_config[k], Path):  # 路径类配置
                    if not v or os.path.isabs(v):
                        v = Path(v)
                    else:
                        v = Path(os.path.abspath(os.path.join(
                            from_dict.get("base_dir") or default_config["base_dir"],
                            v
                        )))
                default[k] = v
        # 特殊处理一下capture相关字段
        def setValue(key: str, value: bool):
            nonlocal default
            # 如果default[key]已经是bool类型, 则不修改原属性
            if isinstance(default.get(key), bool):
                return
            default[key] = value

        for mode in (default["auto_capture"] or "").split(","):
            if mode == "auto":
                setValue("framework_capture", True)
                setValue("error_capture", False)  # teardown时截图就不需要错误截图了
                setValue("assert_capture", True)  # 默认开启
            elif mode == "error":
                setValue("error_capture", True)
                setValue("framework_capture", False)
            elif mode == "assert":
                setValue("assert_capture", True)
        else:
            setValue("framework_capture", False)
            setValue("error_capture", False)
            setValue("assert_capture", False)
            
        for k, v in default.items():
            setattr(self, k, v)
        super(MiniConfig, self).__init__(self.__dict__)

    def __getattr__(self, __k):
        try:
            return self[__k]
        except KeyError:
            return None

    def __setattr__(self, __k, __v):
        if hasattr(self.__class__, __k):
            super().__setattr__(__k, __v)
        else:
            self[__k] = __v

    @property
    def outputs(self):
        return self["outputs"]

    @outputs.setter
    def outputs(self, v):
        if v is None:  # create default outputs dir
            outputs = os.path.join(os.getcwd(), "outputs")
        else:
            outputs = v
        if not os.path.exists(outputs):
            os.makedirs(outputs)
        self["outputs"] = outputs

    @classmethod
    def from_file(cls, filename, encoding=None):
        logger.info("load config from %s", filename)
        _, ext = os.path.splitext(filename)
        for _encoding in set([encoding, "utf8"]):  # 优先使用配置的encoding, 默认使用utf8
            try:
                with open(filename, "r", encoding=_encoding) as f:
                    if ext == ".json":
                        json_dict = json.load(f)
                    elif ext == ".yml" or ext == ".yaml":
                        json_dict = yaml.load(f)
                    else:
                        raise RuntimeError(f"unknown extension {ext} for {filename}")
            except UnicodeDecodeError:
                continue

        if isinstance(json_dict, list):
            config_list = list()
            for c in json_dict:
                if not c.get("base_dir"):
                    c["base_dir"] = os.path.dirname(os.path.abspath(filename))
                config_list.append(MiniConfig(c))
            return config_list
        if not json_dict.get("base_dir"):
            json_dict["base_dir"] = os.path.dirname(os.path.abspath(filename))
        return MiniConfig(json_dict)



if __name__ == "__main__":
    a = MiniConfig({"outputs": "xxxx"})
    print(a.outputs)
    print(a.sss)



# -*- coding: utf-8 -*-
# Created by xiazeng on 2020/3/10
# refer: https://github.com/openstf/minicap
import os.path
from at.core.adbwrap import AdbWrap
from at.core.config import STF_LIB_PATH


class MiniCap:
    CMD = "LD_LIBRARY_PATH=/data/local/tmp /data/local/tmp/minicap"

    def __init__(self, serial=None, display_id=None):
        self.adb = AdbWrap.apply_adb(serial)
        self.display_id = display_id
        self.minicap_path = "/data/local/tmp/minicap"
        self.minicap_so_path = "/data/local/tmp/minicap.so"
        self._check_push()

    def _check_push(self):
        abi = self.adb.get_property("ro.product.cpu.abi")
        minicap_pc_path = os.path.join(STF_LIB_PATH, abi, "minicap").replace("\\", r"\\")
        minicap_so_pc_path = os.path.join(STF_LIB_PATH, abi, "minicap.so").replace("\\", r"\\")

        self.adb.compare_push(minicap_so_pc_path, self.minicap_so_path)
        if self.adb.compare_push(minicap_pc_path, self.minicap_path):
            self.adb.run_shell("chmod 755 %s" % self.minicap_path)

    def get_frame(self):
        w, h = self.adb.screen_size
        projection = "%dx%d@%dx%d/%d" % (h, w, h, w, 0)
        if self.display_id:
            raw_data = self.adb.run_shell(
                self.CMD + " -d " + str(self.display_id) + " -n 'airtest_minicap' -P %s -s" % projection
            )
        else:
            raw_data = self.adb.run_shell(
                self.CMD + " -n 'airtest_minicap' -P %s -s" % projection
            )
        jpg_data = raw_data.split(b"for JPG encoder" + self.adb.line_breaker)[-1]
        jpg_data = jpg_data.replace(self.adb.line_breaker, b"\n")
        return jpg_data
